# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import queue
import re
import threading
import time
from itertools import chain
from typing import Any, Dict, Optional, Tuple, Union

import numpy as np
from PIL import Image

from ....utils import logging
from ....utils.deps import pipeline_requires_extra
from ...common.batch_sampler import ImageBatchSampler
from ...common.reader import ReadImage
from ...utils.benchmark import benchmark
from ...utils.hpi import HPIConfig
from ...utils.pp_option import PaddlePredictorOption
from .._parallel import AutoParallelImageSimpleInferencePipeline
from ..base import BasePipeline
from ..components import CropByBoxes
from ..layout_parsing.merge_table import merge_tables_across_pages
from ..layout_parsing.title_level import assign_levels_to_parsing_res
from ..layout_parsing.utils import construct_img_path, gather_imgs
from .result import BaseResult, PaddleOCRVLBlock, PaddleOCRVLResult
from .uilts import (
    convert_otsl_to_html,
    crop_margin,
    filter_overlap_boxes,
    merge_blocks,
    post_process_for_spotting,
    pre_process_for_spotting,
    tokenize_figure_of_table,
    truncate_repetitive_content,
    untokenize_figure_of_table,
)

IMAGE_LABELS = ["image", "header_image", "footer_image"]


@benchmark.time_methods
class _PaddleOCRVLPipeline(BasePipeline):
    """_PaddleOCRVLPipeline Pipeline"""

    def __init__(
        self,
        config: Dict,
        device: Optional[str] = None,
        pp_option: Optional[PaddlePredictorOption] = None,
        use_hpip: bool = False,
        hpi_config: Optional[Union[Dict[str, Any], HPIConfig]] = None,
        initial_predictor: bool = True,
    ) -> None:
        """
        Initializes the class with given configurations and options.

        Args:
            config (Dict): Configuration dictionary containing various settings.
            device (str, optional): Device to run the predictions on. Defaults to None.
            pp_option (PaddlePredictorOption, optional): PaddlePredictor options. Defaults to None.
            use_hpip (bool, optional): Whether to use the high-performance
                inference plugin (HPIP) by default. Defaults to False.
            hpi_config (Optional[Union[Dict[str, Any], HPIConfig]], optional):
                The default high-performance inference configuration dictionary.
                Defaults to None.
            initial_predictor (bool, optional): Whether to initialize predictors.
        """
        super().__init__(
            device=device, pp_option=pp_option, use_hpip=use_hpip, hpi_config=hpi_config
        )

        if initial_predictor:
            self.use_doc_preprocessor = config.get("use_doc_preprocessor", True)
            if self.use_doc_preprocessor:
                doc_preprocessor_config = config.get("SubPipelines", {}).get(
                    "DocPreprocessor",
                    {
                        "pipeline_config_error": "config error for doc_preprocessor_pipeline!"
                    },
                )
                self.doc_preprocessor_pipeline = self.create_pipeline(
                    doc_preprocessor_config
                )

            self.use_layout_detection = config.get("use_layout_detection", True)
            if self.use_layout_detection:
                layout_det_config = config.get("SubModules", {}).get(
                    "LayoutDetection",
                    {"model_config_error": "config error for layout_det_model!"},
                )
                model_name = layout_det_config.get("model_name", None)
                assert model_name is not None and model_name in [
                    "PP-DocLayoutV2",
                    "PP-DocLayoutV3",
                ], "model_name must be PP-DocLayoutV2 or PP-DocLayoutV3"
                layout_kwargs = {}
                if (threshold := layout_det_config.get("threshold", None)) is not None:
                    layout_kwargs["threshold"] = threshold
                if (
                    layout_nms := layout_det_config.get("layout_nms", None)
                ) is not None:
                    layout_kwargs["layout_nms"] = layout_nms
                if (
                    layout_unclip_ratio := layout_det_config.get(
                        "layout_unclip_ratio", None
                    )
                ) is not None:
                    layout_kwargs["layout_unclip_ratio"] = layout_unclip_ratio
                if (
                    layout_merge_bboxes_mode := layout_det_config.get(
                        "layout_merge_bboxes_mode", None
                    )
                ) is not None:
                    layout_kwargs["layout_merge_bboxes_mode"] = layout_merge_bboxes_mode
                self.layout_det_model = self.create_model(
                    layout_det_config, **layout_kwargs
                )

            self.use_chart_recognition = config.get("use_chart_recognition", False)
            self.use_seal_recognition = config.get("use_seal_recognition", False)

            vl_rec_config = config.get("SubModules", {}).get(
                "VLRecognition",
                {"model_config_error": "config error for vl_rec_model!"},
            )

            self.vl_rec_model = self.create_model(vl_rec_config)
            self.format_block_content = config.get("format_block_content", False)
            self.use_ocr_for_image_block = config.get("use_ocr_for_image_block", False)

            self.batch_sampler = ImageBatchSampler(
                batch_size=config.get("batch_size", 1)
            )
            self.img_reader = ReadImage(format="BGR")
            self.crop_by_boxes = CropByBoxes()

            self.use_queues = config.get("use_queues", False)
            self.merge_layout_blocks = config.get("merge_layout_blocks", True)
            self.markdown_ignore_labels = config.get(
                "markdown_ignore_labels",
                [
                    "number",
                    "footnote",
                    "header",
                    "header_image",
                    "footer",
                    "footer_image",
                    "aside_text",
                ],
            )

    def close(self):
        if hasattr(self, "vl_rec_model"):
            self.vl_rec_model.close()

    def get_model_settings(
        self,
        use_doc_orientation_classify: Union[bool, None],
        use_doc_unwarping: Union[bool, None],
        use_layout_detection: Union[bool, None],
        use_chart_recognition: Union[bool, None],
        use_seal_recognition: Union[bool, None],
        use_ocr_for_image_block: Union[bool, None],
        format_block_content: Union[bool, None],
        merge_layout_blocks: Union[bool, None],
        markdown_ignore_labels: Optional[list[str]] = None,
    ) -> dict:
        """
        Get the model settings based on the provided parameters or default values.

        Args:
            use_doc_orientation_classify (Union[bool, None]): Enables document orientation classification if True. Defaults to system setting if None.
            use_doc_unwarping (Union[bool, None]): Enables document unwarping if True. Defaults to system setting if None.

        Returns:
            dict: A dictionary containing the model settings.

        """
        if use_doc_orientation_classify is None and use_doc_unwarping is None:
            use_doc_preprocessor = self.use_doc_preprocessor
        else:
            if use_doc_orientation_classify is True or use_doc_unwarping is True:
                use_doc_preprocessor = True
            else:
                use_doc_preprocessor = False

        if use_layout_detection is None:
            use_layout_detection = self.use_layout_detection

        if use_chart_recognition is None:
            use_chart_recognition = self.use_chart_recognition

        if use_seal_recognition is None:
            use_seal_recognition = self.use_seal_recognition

        if use_ocr_for_image_block is None:
            use_ocr_for_image_block = self.use_ocr_for_image_block

        if format_block_content is None:
            format_block_content = self.format_block_content

        if merge_layout_blocks is None:
            merge_layout_blocks = self.merge_layout_blocks

        if markdown_ignore_labels is None:
            markdown_ignore_labels = self.markdown_ignore_labels

        return dict(
            use_doc_preprocessor=use_doc_preprocessor,
            use_layout_detection=use_layout_detection,
            use_chart_recognition=use_chart_recognition,
            use_seal_recognition=use_seal_recognition,
            use_ocr_for_image_block=use_ocr_for_image_block,
            format_block_content=format_block_content,
            merge_layout_blocks=merge_layout_blocks,
            markdown_ignore_labels=markdown_ignore_labels,
        )

    def check_model_settings_valid(self, input_params: dict) -> bool:
        """
        Check if the input parameters are valid based on the initialized models.

        Args:
            input_params (Dict): A dictionary containing input parameters.

        Returns:
            bool: True if all required models are initialized according to input parameters, False otherwise.
        """

        if input_params["use_doc_preprocessor"] and not self.use_doc_preprocessor:
            logging.error(
                "Set use_doc_preprocessor, but the models for doc preprocessor are not initialized.",
            )
            return False

        return True

    def get_layout_parsing_results(
        self,
        images,
        layout_det_results,
        imgs_in_doc,
        use_chart_recognition=False,
        use_seal_recognition=False,
        use_ocr_for_image_block=False,
        vlm_kwargs=None,
        merge_layout_blocks=True,
        layout_shape_mode="auto",
    ):
        blocks = []
        has_spotting = False
        drop_figures_set = set()
        min_pixels = vlm_kwargs.pop("min_pixels", None)
        default_min_pixels = min_pixels if min_pixels is not None else 112896
        max_pixels = vlm_kwargs.pop("max_pixels", None)
        default_max_pixels = max_pixels if max_pixels is not None else 1003520

        batch_dict_by_pixel = {}
        id2pixel_key_map = {}
        image_path_to_obj_map = {}
        vis_image_labels = IMAGE_LABELS + ["seal"]
        image_labels = [] if use_ocr_for_image_block else IMAGE_LABELS
        if not use_chart_recognition:
            image_labels += ["chart"]
            vis_image_labels += ["chart"]
        if not use_seal_recognition:
            image_labels += ["seal"]
        for i, (image, layout_det_res, imgs_in_doc_for_img) in enumerate(
            zip(images, layout_det_results, imgs_in_doc)
        ):
            layout_det_res = filter_overlap_boxes(layout_det_res, layout_shape_mode)
            boxes = layout_det_res["boxes"]
            blocks_for_img = self.crop_by_boxes(image, boxes, layout_shape_mode)
            del layout_det_res, boxes
            if merge_layout_blocks:
                blocks_for_img = merge_blocks(
                    blocks_for_img, non_merge_labels=image_labels + ["table"]
                )
            blocks.append(blocks_for_img)
            for j, block in enumerate(blocks_for_img):
                block_img = block["img"]
                block_label = block["label"]
                if block_label not in image_labels and block_img is not None:
                    figure_token_map = {}
                    text_prompt = "OCR:"
                    min_pixels = vlm_kwargs.pop("ocr_min_pixels", default_min_pixels)
                    max_pixels = vlm_kwargs.pop("ocr_max_pixels", default_max_pixels)
                    drop_figures = []
                    if block_label == "table":
                        text_prompt = "Table Recognition:"
                        block_img, figure_token_map, drop_figures = (
                            tokenize_figure_of_table(
                                block_img, block["box"], imgs_in_doc_for_img
                            )
                        )
                        min_pixels = vlm_kwargs.pop(
                            "table_min_pixels", default_min_pixels
                        )
                        max_pixels = vlm_kwargs.pop(
                            "table_max_pixels", default_max_pixels
                        )
                    elif block_label == "chart" and use_chart_recognition:
                        text_prompt = "Chart Recognition:"
                        min_pixels = vlm_kwargs.pop(
                            "chart_min_pixels", default_min_pixels
                        )
                        max_pixels = vlm_kwargs.pop(
                            "chart_max_pixels", default_max_pixels
                        )
                    elif "formula" in block_label and block_label != "formula_number":
                        text_prompt = "Formula Recognition:"
                        crop_img = crop_margin(block_img)
                        w, h, _ = crop_img.shape
                        if w > 2 and h > 2:
                            block_img = crop_img
                        min_pixels = vlm_kwargs.pop(
                            "formula_min_pixels", default_min_pixels
                        )
                        max_pixels = vlm_kwargs.pop(
                            "formula_max_pixels", default_max_pixels
                        )
                    elif block_label == "spotting":
                        text_prompt = "Spotting:"
                        has_spotting = True
                        min_pixels = 112896
                        max_pixels = 1605632
                        block_img = pre_process_for_spotting(block_img)
                    elif block_label == "seal" and use_seal_recognition:
                        text_prompt = "Seal Recognition:"
                        min_pixels = vlm_kwargs.pop(
                            "seal_min_pixels", default_min_pixels
                        )
                        max_pixels = vlm_kwargs.pop(
                            "seal_max_pixels", default_max_pixels
                        )
                    pixel_key = (min_pixels, max_pixels)
                    if pixel_key not in batch_dict_by_pixel:
                        batch_dict_by_pixel[pixel_key] = {
                            "images": [],
                            "queries": [],
                            "figure_token_maps": [],
                            "vlm_block_ids": [],
                            "curr_vlm_block_idx": 0,
                        }
                    batch_dict_by_pixel[pixel_key]["images"].append(block_img)
                    batch_dict_by_pixel[pixel_key]["queries"].append(text_prompt)
                    batch_dict_by_pixel[pixel_key]["figure_token_maps"].append(
                        figure_token_map
                    )
                    batch_dict_by_pixel[pixel_key]["vlm_block_ids"].append((i, j))
                    id2pixel_key_map[(i, j)] = pixel_key
                    drop_figures_set.update(drop_figures)
            del blocks_for_img
        del images, layout_det_results

        if vlm_kwargs is None:
            vlm_kwargs = {}
        elif vlm_kwargs.get("max_new_tokens", None) is None:
            vlm_kwargs["max_new_tokens"] = 4096

        for pixel_key in batch_dict_by_pixel:
            min_pixels, max_pixels = pixel_key
            kwargs = {
                "use_cache": True,
                "min_pixels": min_pixels,
                "max_pixels": max_pixels,
                **vlm_kwargs,
            }
            images = batch_dict_by_pixel[pixel_key]["images"]
            queries = batch_dict_by_pixel[pixel_key]["queries"]
            batch_results = list(
                self.vl_rec_model.predict(
                    [
                        {
                            "image": image,
                            "query": query,
                        }
                        for image, query in zip(images, queries)
                    ],
                    skip_special_tokens=False if has_spotting else True,
                    **kwargs,
                )
            )
            del images, queries
            batch_dict_by_pixel[pixel_key]["vlm_results"] = batch_results

        parsing_res_lists = []
        table_res_lists = []
        spotting_res_list = []
        table_blocks = []
        for i, blocks_for_img in enumerate(blocks):
            parsing_res_list = []
            table_res_list = []
            spotting_res = {}
            for j, block in enumerate(blocks_for_img):
                block_img = block["img"]
                block_bbox = block["box"]
                block_label = block["label"]
                block_content = ""
                figure_token_map = {}
                if (i, j) in id2pixel_key_map:
                    pixel_key = id2pixel_key_map[(i, j)]
                    pixel_info = batch_dict_by_pixel[pixel_key]
                    curr_vlm_block_idx = pixel_info["curr_vlm_block_idx"]
                    assert curr_vlm_block_idx < len(
                        pixel_info["vlm_block_ids"]
                    ) and pixel_info["vlm_block_ids"][curr_vlm_block_idx] == (i, j)
                    vl_rec_result = pixel_info["vlm_results"][curr_vlm_block_idx]
                    block_img4vl = pixel_info["images"][curr_vlm_block_idx]
                    figure_token_map = pixel_info["figure_token_maps"][
                        curr_vlm_block_idx
                    ]
                    curr_vlm_block_idx += 1
                    pixel_info["curr_vlm_block_idx"] = curr_vlm_block_idx
                    vl_rec_result["image"] = block_img4vl
                    result_str = vl_rec_result.get("result", "")
                    if result_str is None:
                        result_str = ""
                    min_count = 5000 if block_label == "table" else 50
                    result_str = truncate_repetitive_content(
                        result_str, min_count=min_count
                    )
                    if ("\\(" in result_str and "\\)" in result_str) or (
                        "\\[" in result_str and "\\]" in result_str
                    ):
                        result_str = result_str.replace("$", "")

                        result_str = (
                            result_str.replace("\\(", " $ ")
                            .replace("\\)", " $")
                            .replace("\\[\\[", "\\[")
                            .replace("\\]\\]", "\\]")
                            .replace("\\[", " $$ ")
                            .replace("\\]", " $$ ")
                        )
                        if block_label == "formula_number":
                            result_str = result_str.replace("$", "")
                    if block_label == "table":
                        html_str = convert_otsl_to_html(result_str)
                        if html_str != "":
                            result_str = html_str
                    if block_label == "spotting":
                        h, w = block_img.shape[:2]
                        result_str, spotting_res = post_process_for_spotting(
                            result_str, w, h
                        )

                    block_content = result_str
                block_info = PaddleOCRVLBlock(
                    label=block_label,
                    bbox=block_bbox,
                    content=block_content,
                    group_id=block.get("group_id", None),
                    polygon_points=block.get("polygon_points", None),
                )
                if block_label == "table":
                    table_blocks.append(
                        {
                            "figure_token_map": figure_token_map,
                            "block": block_info,
                        }
                    )
                if block_label in vis_image_labels and block_img is not None:
                    img_path = construct_img_path(block["label"], block["box"])
                    image_path_to_obj_map[img_path] = block_info
                    if img_path not in drop_figures_set:
                        import cv2

                        block_img = cv2.cvtColor(block_img, cv2.COLOR_BGR2RGB)
                        block_info.image = {
                            "path": img_path,
                            "img": Image.fromarray(block_img),
                        }
                    else:
                        continue

                parsing_res_list.append(block_info)
                del block_info, block_img
            # TODO(changdazhou): append table res to table_res_list
            for blk_info in table_blocks:
                block = blk_info["block"]
                figure_token_map = blk_info["figure_token_map"]
                block.content = untokenize_figure_of_table(
                    block.content, figure_token_map, image_path_to_obj_map
                )
            parsing_res_lists.append(parsing_res_list)
            table_res_lists.append(table_res_list)
            spotting_res_list.append(spotting_res)
            del parsing_res_list, table_res_list, spotting_res

        return (
            parsing_res_lists,
            table_res_lists,
            spotting_res_list,
            imgs_in_doc,
        )

    def predict(
        self,
        input: Union[str, list[str], np.ndarray, list[np.ndarray]],
        use_doc_orientation_classify: Union[bool, None] = False,
        use_doc_unwarping: Union[bool, None] = False,
        use_layout_detection: Union[bool, None] = None,
        use_chart_recognition: Union[bool, None] = None,
        use_seal_recognition: Union[bool, None] = None,
        use_ocr_for_image_block: Union[bool, None] = None,
        layout_threshold: Optional[Union[float, dict]] = None,
        layout_nms: Optional[bool] = None,
        layout_unclip_ratio: Optional[Union[float, Tuple[float, float], dict]] = None,
        layout_merge_bboxes_mode: Optional[str] = None,
        layout_shape_mode: Optional[str] = "auto",
        use_queues: Optional[bool] = None,
        prompt_label: Optional[Union[str, None]] = None,
        format_block_content: Union[bool, None] = None,
        repetition_penalty: Optional[float] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        min_pixels: Optional[int] = None,
        max_pixels: Optional[int] = None,
        max_new_tokens: Optional[int] = None,
        merge_layout_blocks: Optional[bool] = None,
        markdown_ignore_labels: Optional[list[str]] = None,
        vlm_extra_args: Optional[dict] = None,
        **kwargs,
    ) -> PaddleOCRVLResult:
        """
        Predicts the layout parsing result for the given input.

        Args:
            input (Union[str, list[str], np.ndarray, list[np.ndarray]]): Input image path, list of image paths,
                                                                        numpy array of an image, or list of numpy arrays.
            use_doc_orientation_classify (Optional[bool]): Whether to use document orientation classification.
            use_doc_unwarping (Optional[bool]): Whether to use document unwarping.
            use_layout_detection (Optional[bool]): Whether to use layout detection. Default is None.
            use_chart_recognition (Optional[bool]): Whether to use chart recognition. Default is None.
            use_seal_recognition (Optional[bool]): Whether to use seal recognition. Default is None.
            layout_threshold (Optional[float]): The threshold value to filter out low-confidence predictions. Default is None.
            layout_nms (bool, optional): Whether to use layout-aware NMS. Defaults to False.
            layout_unclip_ratio (Optional[Union[float, Tuple[float, float]]], optional): The ratio of unclipping the bounding box.
                Defaults to None.
                If it's a single number, then both width and height are used.
                If it's a tuple of two numbers, then they are used separately for width and height respectively.
                If it's None, then no unclipping will be performed.
            layout_merge_bboxes_mode (Optional[str], optional): The mode for merging bounding boxes. Defaults to None.
            layout_shape_mode (Optional[str], optional): The mode for layout shape. Defaults to "auto", [ "rect", "quad","poly", "auto"] are supported.
            use_queues (Optional[bool], optional): Whether to use queues. Defaults to None.
            prompt_label (Optional[Union[str, None]], optional): The label of the prompt in ['ocr', 'formula', 'table', 'chart']. Defaults to None.
            format_block_content (Optional[bool]): Whether to format the block content. Default is None.
            repetition_penalty (Optional[float]): The repetition penalty parameter used for VL model sampling. Default is None.
            temperature (Optional[float]): Temperature parameter used for VL model sampling. Default is None.
            top_p (Optional[float]): Top-p parameter used for VL model sampling. Default is None.
            min_pixels (Optional[int]): The minimum number of pixels allowed when the VL model preprocesses images. Default is None.
            max_pixels (Optional[int]): The maximum number of pixels allowed when the VL model preprocesses images. Default is None.
            max_new_tokens (Optional[int]): The maximum number of new tokens. Default is None.
            merge_layout_blocks (Optional[bool]): Whether to merge layout blocks. Default is None.
            markdown_ignore_labels (Optional[list[str]]): The list of ignored markdown labels. Default is None.
            **kwargs (Any): Additional settings to extend functionality.

        Returns:
            PaddleOCRVLResult: The predicted layout parsing result.
        """
        model_settings = self.get_model_settings(
            use_doc_orientation_classify,
            use_doc_unwarping,
            use_layout_detection,
            use_chart_recognition,
            use_seal_recognition,
            use_ocr_for_image_block,
            format_block_content,
            merge_layout_blocks,
            markdown_ignore_labels,
        )

        model_settings["return_layout_polygon_points"] = (
            False if layout_shape_mode == "rect" else True
        )

        if not self.check_model_settings_valid(model_settings):
            yield {"error": "the input params for model settings are invalid!"}

        if use_queues is None:
            use_queues = self.use_queues

        if vlm_extra_args is None:
            vlm_extra_args = {}

        if not model_settings["use_layout_detection"]:
            prompt_label = prompt_label if prompt_label else "ocr"
            if prompt_label.lower() == "chart":
                model_settings["use_chart_recognition"] = True
            elif prompt_label.lower() == "seal":
                model_settings["use_seal_recognition"] = True
            assert prompt_label.lower() in [
                "ocr",
                "formula",
                "table",
                "chart",
                "spotting",
                "seal",
            ], f"Layout detection is disabled (use_layout_detection=False). 'prompt_label' must be one of ['ocr', 'formula', 'table', 'chart'], but got '{prompt_label}'."

        def _process_cv(batch_data, new_batch_size=None):
            if not new_batch_size:
                new_batch_size = len(batch_data)

            for idx in range(0, len(batch_data), new_batch_size):
                instances = batch_data.instances[idx : idx + new_batch_size]
                input_paths = batch_data.input_paths[idx : idx + new_batch_size]
                page_indexes = batch_data.page_indexes[idx : idx + new_batch_size]
                page_counts = batch_data.page_counts[idx : idx + new_batch_size]

                image_arrays = self.img_reader(instances)

                if model_settings["use_doc_preprocessor"]:
                    doc_preprocessor_results = list(
                        self.doc_preprocessor_pipeline(
                            image_arrays,
                            use_doc_orientation_classify=use_doc_orientation_classify,
                            use_doc_unwarping=use_doc_unwarping,
                        )
                    )
                else:
                    doc_preprocessor_results = [
                        {"output_img": arr} for arr in image_arrays
                    ]

                doc_preprocessor_images = [
                    item["output_img"] for item in doc_preprocessor_results
                ]
                if model_settings["use_layout_detection"]:
                    layout_det_results = list(
                        self.layout_det_model(
                            doc_preprocessor_images,
                            threshold=layout_threshold,
                            layout_nms=layout_nms,
                            layout_unclip_ratio=layout_unclip_ratio,
                            layout_merge_bboxes_mode=layout_merge_bboxes_mode,
                            layout_shape_mode=layout_shape_mode,
                            filter_overlap_boxes=False,
                        )
                    )

                    imgs_in_doc = [
                        gather_imgs(doc_pp_img, layout_det_res["boxes"])
                        for doc_pp_img, layout_det_res in zip(
                            doc_preprocessor_images, layout_det_results
                        )
                    ]
                else:
                    layout_det_results = []
                    for doc_preprocessor_image in doc_preprocessor_images:
                        layout_det_results.append(
                            {
                                "input_path": None,
                                "page_index": None,
                                "boxes": [
                                    {
                                        "cls_id": 0,
                                        "label": prompt_label.lower(),
                                        "score": 1,
                                        "coordinate": [
                                            0,
                                            0,
                                            doc_preprocessor_image.shape[1],
                                            doc_preprocessor_image.shape[0],
                                        ],
                                    }
                                ],
                            }
                        )
                    imgs_in_doc = [[] for _ in layout_det_results]

                yield input_paths, page_indexes, page_counts, doc_preprocessor_images, doc_preprocessor_results, layout_det_results, imgs_in_doc

        def _process_vlm(results_cv):
            (
                input_paths,
                page_indexes,
                page_counts,
                doc_preprocessor_images,
                doc_preprocessor_results,
                layout_det_results,
                imgs_in_doc,
            ) = results_cv

            (
                parsing_res_lists,
                table_res_lists,
                spotting_res_list,
                imgs_in_doc,
            ) = self.get_layout_parsing_results(
                images=doc_preprocessor_images,
                layout_det_results=layout_det_results,
                imgs_in_doc=imgs_in_doc,
                use_chart_recognition=model_settings["use_chart_recognition"],
                use_seal_recognition=model_settings["use_seal_recognition"],
                use_ocr_for_image_block=model_settings["use_ocr_for_image_block"],
                vlm_kwargs={
                    "repetition_penalty": repetition_penalty,
                    "temperature": temperature,
                    "top_p": top_p,
                    "min_pixels": min_pixels,
                    "max_pixels": max_pixels,
                    "max_new_tokens": max_new_tokens,
                    **vlm_extra_args,
                },
                merge_layout_blocks=model_settings["merge_layout_blocks"],
                layout_shape_mode=layout_shape_mode,
            )

            for (
                input_path,
                page_index,
                page_count,
                doc_preprocessor_image,
                doc_preprocessor_res,
                layout_det_res,
                table_res_list,
                parsing_res_list,
                spotting_res,
                imgs_in_doc_for_img,
            ) in zip(
                input_paths,
                page_indexes,
                page_counts,
                doc_preprocessor_images,
                doc_preprocessor_results,
                layout_det_results,
                table_res_lists,
                parsing_res_lists,
                spotting_res_list,
                imgs_in_doc,
            ):
                single_img_res = {
                    "input_path": input_path,
                    "page_index": page_index,
                    "page_count": page_count,
                    "width": doc_preprocessor_image.shape[1],
                    "height": doc_preprocessor_image.shape[0],
                    "doc_preprocessor_res": doc_preprocessor_res,
                    "layout_det_res": layout_det_res,
                    "table_res_list": table_res_list,
                    "parsing_res_list": parsing_res_list,
                    "spotting_res": spotting_res,
                    "imgs_in_doc": imgs_in_doc_for_img,
                    "model_settings": model_settings,
                }
                yield PaddleOCRVLResult(single_img_res)

        if use_queues:
            max_num_batches_in_process = 64
            queue_input = queue.Queue(maxsize=max_num_batches_in_process)
            queue_cv = queue.Queue(maxsize=max_num_batches_in_process)
            queue_vlm = queue.Queue(
                maxsize=self.batch_sampler.batch_size * max_num_batches_in_process
            )
            event_shutdown = threading.Event()
            event_data_loading_done = threading.Event()
            event_cv_processing_done = threading.Event()
            event_vlm_processing_done = threading.Event()

            def _worker_input(input_):
                all_batch_data = self.batch_sampler(input_)
                while not event_shutdown.is_set():
                    try:
                        batch_data = next(all_batch_data)
                    except StopIteration:
                        break
                    except Exception as e:
                        queue_input.put((False, "input", e))
                        break
                    else:
                        queue_input.put((True, batch_data))
                        del batch_data
                event_data_loading_done.set()

            def _worker_cv():
                while not event_shutdown.is_set():
                    try:
                        item = queue_input.get(timeout=0.5)
                    except queue.Empty:
                        if event_data_loading_done.is_set():
                            event_cv_processing_done.set()
                            break
                        continue
                    if not item[0]:
                        queue_cv.put(item)
                        break
                    try:
                        for results_cv in _process_cv(
                            item[1],
                            (
                                self.layout_det_model.batch_sampler.batch_size
                                if model_settings["use_layout_detection"]
                                else None
                            ),
                        ):
                            queue_cv.put((True, results_cv))
                            del results_cv
                        del item
                    except Exception as e:
                        queue_cv.put((False, "cv", e))
                        break

            def _worker_vlm():
                MAX_QUEUE_DELAY_SECS = 0.5
                MAX_NUM_BOXES = self.vl_rec_model.batch_sampler.batch_size

                while not event_shutdown.is_set():
                    results_cv_list = []
                    start_time = time.time()
                    should_break = False
                    num_boxes = 0
                    while True:
                        remaining_time = MAX_QUEUE_DELAY_SECS - (
                            time.time() - start_time
                        )
                        if remaining_time <= 0:
                            break
                        try:
                            item = queue_cv.get(timeout=remaining_time)
                        except queue.Empty:
                            break
                        if not item[0]:
                            queue_vlm.put(item)
                            should_break = True
                            break
                        results_cv_list.append(item[1])
                        del item
                        for res in results_cv_list[-1][5]:
                            num_boxes += len(res["boxes"])
                        if num_boxes >= MAX_NUM_BOXES:
                            break
                    if should_break:
                        break
                    if not results_cv_list:
                        if event_cv_processing_done.is_set():
                            event_vlm_processing_done.set()
                            break
                        continue

                    merged_results_cv = [
                        list(chain.from_iterable(lists))
                        for lists in zip(*results_cv_list)
                    ]
                    del results_cv_list

                    try:
                        for result_vlm in _process_vlm(merged_results_cv):
                            queue_vlm.put((True, result_vlm))
                            del result_vlm
                        del merged_results_cv
                    except Exception as e:
                        queue_vlm.put((False, "vlm", e))
                        break

            thread_input = threading.Thread(
                target=_worker_input, args=(input,), daemon=False
            )
            thread_input.start()
            thread_cv = threading.Thread(target=_worker_cv, daemon=False)
            thread_cv.start()
            thread_vlm = threading.Thread(target=_worker_vlm, daemon=False)
            thread_vlm.start()

        try:
            if use_queues:
                while not (event_vlm_processing_done.is_set() and queue_vlm.empty()):
                    try:
                        item = queue_vlm.get(timeout=0.5)
                    except queue.Empty:
                        if event_vlm_processing_done.is_set():
                            break
                        continue
                    if not item[0]:
                        raise RuntimeError(
                            f"Exception from the '{item[1]}' worker: {item[2]}"
                        )
                    else:
                        yield item[1]
            else:
                for batch_data in self.batch_sampler(input):
                    results_cv_list = list(_process_cv(batch_data))
                    assert len(results_cv_list) == 1, len(results_cv_list)
                    results_cv = results_cv_list[0]
                    for res in _process_vlm(results_cv):
                        yield res
                    del res, results_cv, results_cv_list, batch_data
        finally:
            if use_queues:
                event_shutdown.set()
                thread_input.join(timeout=5)
                if thread_input.is_alive():
                    logging.warning("Input worker did not terminate in time")
                thread_cv.join(timeout=5)
                if thread_cv.is_alive():
                    logging.warning("CV worker did not terminate in time")
                thread_vlm.join(timeout=5)
                if thread_vlm.is_alive():
                    logging.warning("VLM worker did not terminate in time")

    def concatenate_markdown_pages(self, markdown_list: list) -> tuple:
        """
        Concatenate Markdown content from multiple pages into a single document.

        Args:
            markdown_list (list): A list containing Markdown data for each page.

        Returns:
            tuple: A tuple containing the processed Markdown text.
        """
        markdown_texts = ""

        for res in markdown_list:
            markdown_texts += "\n\n" + res["markdown_texts"]

        return markdown_texts

    def concatenate_pages(
        self,
        res_list: list,
        merge_table: bool = True,
        title_level: bool = True,
        merge_pages: bool = False,
    ):
        """Concatenate layout parsing results from multiple pages.

        Args:
            res_list: List of page parsing results
            merge_talble: Whether to merge tables across pages
            title_level: Whether to assign title levels
            merge_pages: Whether to concatenate pages using the new consolidate_pages() logic

        Returns:
            PaddleOCRVLResult: Combined OCR-VL result after merge_table or title_level policy
        """
        logging.warning(
            f"DeprecationWarning: `concatenate_pages()` is deprecated as of v3.3.14 and will be removed in v3.4.0. Please use `restructure_pages()` instead. It provides better support for table merging and title restructuring."
        )
        return self.restructure_pages(res_list, merge_table, title_level, merge_pages)

    def restructure_pages(
        self,
        res_list: list,
        merge_tables: bool = True,
        relevel_titles: bool = True,
        concatenate_pages: bool = False,
    ):
        """Restructure layout parsing results from multiple pages.
        Args:
            res_list: List of page parsing results
            merge_tables: Whether to merge tables across pages
            relevel_titles: Whether to relevel titles
            concatenate_pages: Whether to concatenate pages to a single document

        Returns:
            PaddleOCRVLResult: Combined OCR-VL result after merge_tables or relevel_titles policy
        """

        if len(res_list) == 0:
            return []

        def _get_img_obj(block, model_settings):
            if block.get("image", None):
                return block["image"]
            if block["block_label"] in ("image", "seal") or (
                block["block_label"] == "chart"
                and not model_settings.get("use_chart_recognition", False)
            ):
                path = construct_img_path(block["block_label"], block["block_bbox"])
                return {"path": path, "img": None}
            return None

        def _conver_blocks_to_obj(blocks, model_settings):
            res = []
            for block in blocks:
                obj = PaddleOCRVLBlock(
                    label=block["block_label"],
                    bbox=block["block_bbox"],
                    polygon_points=block.get("block_polygon_points", None),
                    content=re.sub(r"^#+\s", "", block["block_content"]),
                    group_id=block.get("group_id", None),
                )
                if img := _get_img_obj(block, model_settings):
                    obj.image = img
                res.append(obj)
            return res

        global_block_id = 0
        obj_res_list = []
        for one_page_res in res_list:
            if not isinstance(one_page_res, BaseResult):
                one_page_res = one_page_res["res"]
                one_page_res["imgs_in_doc"] = []
                blocks = one_page_res.get("parsing_res_list", [])
                model_settings = one_page_res.get("model_settings", {})
                blocks = _conver_blocks_to_obj(blocks, model_settings)
            else:
                blocks = one_page_res["parsing_res_list"]
                model_settings = one_page_res.get("model_settings", {})
            parsing_res_list = []
            for block in blocks:
                block.global_block_id = global_block_id
                block.global_group_id = global_block_id
                global_block_id += 1
                parsing_res_list.append(block)

            one_page_res["parsing_res_list"] = parsing_res_list
            obj_res_list.append(one_page_res)
        res_list = obj_res_list

        blocks_by_page = [res["parsing_res_list"] for res in res_list]

        if merge_tables:
            blocks_by_page = merge_tables_across_pages(blocks_by_page)
        if relevel_titles:
            blocks_by_page = assign_levels_to_parsing_res(blocks_by_page)

        concatenate_res = []
        if concatenate_pages:
            all_page_res = res_list[0]
            all_page_res["parsing_res_list"] = [
                blk for blks in blocks_by_page for blk in blks
            ]
            all_page_res["page_index"] = None
            all_page_res["page_count"] = len(res_list)
            if model_settings["use_layout_detection"]:
                all_page_res["layout_det_res"] = [
                    res["layout_det_res"] for res in res_list
                ]
            if model_settings["use_doc_preprocessor"]:
                all_page_res["doc_preprocessor_res"] = [
                    res["doc_preprocessor_res"] for res in res_list
                ]
            concatenate_res.append(PaddleOCRVLResult(all_page_res))
        else:
            for page_idx, one_page_res in enumerate(res_list):
                one_page_res["parsing_res_list"] = blocks_by_page[page_idx]
                concatenate_res.append(PaddleOCRVLResult(one_page_res))

        yield from concatenate_res


class _BasePaddleOCRVLPipeline(AutoParallelImageSimpleInferencePipeline):
    @property
    def _pipeline_cls(self):
        return _PaddleOCRVLPipeline

    def _get_batch_size(self, config):
        return config.get("batch_size", 1)


@pipeline_requires_extra("ocr")
class PaddleOCRVLPipeline(_BasePaddleOCRVLPipeline):
    entities = "PaddleOCR-VL"


@pipeline_requires_extra("ocr")
class PaddleOCRVL15Pipeline(_BasePaddleOCRVLPipeline):
    entities = "PaddleOCR-VL-1.5"



# Copyright (c) 2025 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

from ....utils.deps import require_genai_engine_plugin
from ..configs.utils import (
    backend_config_to_args,
    set_config_defaults,
    update_backend_config,
)


def run_fastdeploy_server(
    host, port, model_name, model_dir, config, chat_template_path
):
    require_genai_engine_plugin("fastdeploy-server")

    if chat_template_path:
        set_config_defaults(config, {"chat-template": str(chat_template_path)})

    update_backend_config(
        config,
        {
            "model": model_dir,
            "host": host,
            "port": port,
        },
    )

    args = backend_config_to_args(config)
    sys.argv[1:] = args

    from fastdeploy.entrypoints.openai.api_server import main as run

    run()



# Copyright (c) 2025 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Common distributed utils.
"""

import paddle
import paddle.nn.functional as F
from paddle import distributed as dist
from paddle.autograd import PyLayer
from paddle.distributed import fleet
from paddle.distributed.fleet.meta_parallel import (
    ColumnParallelLinear,
    RowParallelLinear,
)
from paddle.distributed.fleet.utils.sequence_parallel_utils import (
    AllGatherOp,
    ColumnSequenceParallelLinear,
    GatherOp,
    ReduceScatterOp,
    RowSequenceParallelLinear,
    ScatterOp,
    all_gather,
    mark_as_sequence_parallel_parameter,
    scatter,
)
from paddle.incubate.tensor.manipulation import create_async_load

from .._refined_recompute.utils import RefinedRecomputeFunction

__all__ = [
    "get_hcg",
    "_parallel_matmul",
    "scatter_axis",
    "mp_slice",
    "all_gather_varlen",
    "ColumnParallelLinear",
    "ColumnSequenceParallelLinear",
    "RowParallelLinear",
    "RowSequenceParallelLinear",
    "GatherOp",
    "ScatterOp",
    "mark_as_sequence_parallel_parameter",
    "RRColumnSequenceParallelLinear",
    "RRRowSequenceParallelLinear",
    "AllGatherVarlenOp",
    "sequence_parallel_sparse_mask_labels",
    "get_async_loader",
    "hack_offload_wait",
    "hack_reload_wait",
    "all_gather_group",
    "reduce_scatter_group",
]


def get_hcg():
    """
    Get hybrid communicate group.
    """
    return fleet.get_hybrid_communicate_group()


def _parallel_matmul(
    x,
    y,
    bias=None,
    transpose_y=False,
    tensor_parallel_degree=1,
    tensor_parallel_output=True,
    fuse_linear=False,
):
    """
    Performs parallel matrix multiplication with tensor model parallelism support.

    Args:
        x (paddle.Tensor): Input tensor with shape [batch_size, seq_len, hidden_size]
        y (Union[paddle.Tensor, EagerParamBase]): Weight matrix which can be:
            - Regular tensor
            - Distributed parameter in tensor parallel mode
        bias (Optional[paddle.Tensor]): Optional bias tensor
        transpose_y (bool): Whether to transpose the 'y' matrix before multiplication
        tensor_parallel_degree (int): Degree of tensor model parallelism (default: 1)
        tensor_parallel_output (bool): Whether to keep output in tensor parallel format
            or gather across devices (default: True)
        fuse_linear (bool): Whether to use fused linear operation for optimization

    Returns:
        paddle.Tensor

    Raises:
        AssertionError: If tensor parallel is enabled but weight is not distributed
        AttributeError: If called without distributed.launch context
    """
    if tensor_parallel_degree > 1:
        if isinstance(y, paddle.base.framework.EagerParamBase):
            assert y.is_distributed
        # if not running under distributed.launch, it will raise AttributeError: 'Fleet' object has no attribute '_hcg'
        pg = fleet.get_hybrid_communicate_group().get_model_parallel_group()
        input_parallel = paddle.distributed.collective._c_identity(x, group=pg)

        if transpose_y:
            logits = paddle.matmul(input_parallel, y, transpose_y=True)
            if bias is not None:
                logits += bias
        else:
            if fuse_linear:
                logits = paddle.incubate.nn.functional.fused_linear(
                    input_parallel, y, bias
                )
            else:
                logits = F.linear(input_parallel, y, bias)

        if tensor_parallel_output:
            return logits

        return paddle.distributed.collective._c_concat(logits, group=pg)

    else:
        if fuse_linear:
            logits = paddle.incubate.nn.functional.fused_linear(
                x, y, bias, transpose_weight=transpose_y
            )
        else:
            logits = paddle.matmul(x, y, transpose_y=transpose_y)
            if bias is not None:
                logits += bias
        return logits


def scatter_axis(input, group=None, axis=0):
    """
    Uniformly splits the `input` along dimension 0 across model parallel groups.
    This API is not related to `distributed.scatter`.

    Args:
        input: Input tensor to be split
        group: Communication group for parallel processing (default: model parallel group)
        axis: Dimension along which to split (default: 0)

    Returns:
        A slice of the input tensor corresponding to this rank's portion
    """
    if group is None:
        hcg = fleet.get_hybrid_communicate_group()
        group = hcg.get_model_parallel_group()
    parallelism = group.nranks
    if parallelism == 1:
        return input.clone()
    rank = group.rank
    seq_len = input.shape[axis]
    assert seq_len % parallelism == 0, (
        f"Input sequence length {seq_len} can't be divided exactly"
        f" by sequence parallelism {parallelism}"
    )
    interval = seq_len // parallelism
    input = paddle.slice(
        input, axes=[axis], starts=[interval * rank], ends=[interval * (rank + 1)]
    )
    # slice uses stride, so we maintain the memory of whole input, use assign to free the whole input
    # which can avoid OOM.
    input = paddle.assign(input)
    return input


def mp_slice(x, indices=None, group=None, axis=0):
    """
    Slices tensor `x` along dimension 0 according to `indices` without communication.

    Args:
        x: Input tensor to be sliced
        indices: List of indices defining how to slice the tensor
        group: Communication group for parallel processing (default: model parallel group)
        axis: Dimension along which to slice (default: 0)

    Returns:
        A slice of the input tensor corresponding to this rank's portion
    """
    if indices is None:
        return scatter(x, group, axis)
    if group is None:
        hcg = fleet.get_hybrid_communicate_group()
        group = hcg.get_model_parallel_group()
    parallelism = group.nranks
    if parallelism == 1:
        return x
    rank = group.rank
    assert len(indices) == parallelism, (len(indices), parallelism)
    indices = F.pad(paddle.to_tensor(indices).cumsum(0), [1, 0])
    input = paddle.slice(
        x, axes=[axis], starts=[indices[rank]], ends=[indices[rank + 1]]
    )
    input = paddle.assign(input)
    return input


def all_gather_varlen(input, indices, group=None, axis=0, sync_op=True):
    """
    Variable-length version of `all_gather` that behaves similarly to `distributed.all_gather`.

    Args:
        input: Local tensor to be gathered
        indices: List of sizes from each rank indicating how much to gather from each
        group: Communication group for parallel processing (default: model parallel group)
        axis: Dimension along which to gather (only 0 is supported)
        sync_op: Whether to synchronize the operation

    Returns:
        A concatenated tensor containing all gathered data
    """
    assert axis == 0, "only support axis=0"
    if group is None:
        hcg = fleet.get_hybrid_communicate_group()
        group = hcg.get_model_parallel_group()
    parallelism = group.nranks
    input_sizes = [len(input)] * parallelism
    output_sizes = indices
    out = paddle.empty([sum(indices)] + input.shape[1:], dtype=input.dtype)
    task = dist.stream.alltoall_single(
        out,
        (
            paddle.concat([input] * parallelism, 0) if len(input) else input
        ),  # TODO: check this
        output_sizes,  # input-size
        input_sizes,
        group=group,
        sync_op=sync_op,
        use_calc_stream=sync_op,
    )
    task.wait()
    return out


class ReduceScatterGroupOp(PyLayer):
    """
    Perform group reduce scatter.
    """

    @staticmethod
    def forward(ctx, input, group=None):
        """Forward pass: Reduce-Scatter operation
        Args:
            input (Tensor):  Input tensor with shape [s, b, h].
                            The 's' dimension will be split across model parallel group.
            group (ProcessGroup): Model parallel process group,
                                uses global group by default.
        Returns:
            Tensor: Output tensor after Reduce-Scatter with shape [s/n, b, h],
                   each device holds partial data of the original input.
        """
        ctx.group = group
        return reduce_scatter_group(input, group=group)

    @staticmethod
    def backward(ctx, grad):
        """Backward pass: All-Gather operation
        Args:
            grad (Tensor): Upstream gradient with shape [s/n, b, h]
        Returns:
            Tensor: Full gradient after All-Gather with restored shape [s, b, h],
                   aggregating gradients from all devices in model parallel group.
        """
        return all_gather_group(grad, group=ctx.group)


class AllGatherGroupOp(PyLayer):
    """
    Perform group allgather.
    """

    @staticmethod
    def forward(ctx, input, group=None):
        """Forward pass: All-Gather operation
        Args:
            input (Tensor):  Partitioned tensor with shape [s/n, b, h]
                            The 's' dimension is distributed across devices
            group (ProcessGroup): Model parallel process group,
                                uses global group by default
        Returns:
            Tensor: Assembled tensor after All-Gather with shape [s, b, h],
                   containing full parameter from all devices
        """
        ctx.group = group
        return all_gather_group(input, group=group)

    @staticmethod
    def backward(ctx, grad):
        """Backward pass: Reduce-Scatter operation
        Args:
            grad (Tensor): Full gradient tensor with shape [s, b, h]
        Returns:
            Tensor: Scattered gradient with shape [s/n, b, h],
                   distributing reduced gradients to each device
        """
        return reduce_scatter_group(grad, group=ctx.group)


class RRColumnSequenceParallelLinear(ColumnSequenceParallelLinear):
    """
    ColumnSequenceParallelLinear with refined recompute.
    """

    def __init__(
        self,
        in_features,
        out_features,
        weight_attr=None,
        has_bias=None,
        gather_output=True,
        fuse_matmul_bias=False,
        mp_group=None,
        use_rr=False,
        name=None,
    ):
        """
        Initializes a ColumnSequenceParallelLinear module.

        Args:
            in_features (int): The number of input features.
            out_features (int): The number of output features.
            weight_attr (ParamAttr, optional): The parameter attribute for the learnable
                weight matrix. Default: None.
            has_bias (bool, optional): Whether the layer uses a bias. By default, it is set to False.
                If ``has_bias`` is set to False, no bias term is used. If ``has_bias`` is set to True,
                a bias vector is used. Default: None, which means inherit the value of `has_bias`
                from the current instance's `has_bias`.
            gather_output (bool, optional): Whether to gather all outputs from all ranks during forward pass.
                Default: True. If True, all outputs from all ranks are gathered during forward pass, which
                makes sure that each example's output is produced only once. If False, all outputs are
                produced on each rank separately, and the outputs from different ranks may overlap.
                This can save communication time but may cause slower convergence. Default: True.
            fuse_matmul_bias (bool, optional): Whether to fuse matmul and bias into one op. Default: False.
            mp_group (paddle.distributed.Group, optional): The group for model parallel. Default: None.
            use_rr (bool, optional): Whether to use refined rcompute. Default: False.
            name (str, optional): Name for the instance to use in tracebacks. Default: None.
        """
        super().__init__(
            in_features=in_features,
            out_features=out_features,
            weight_attr=weight_attr,
            has_bias=has_bias,
            gather_output=gather_output,
            fuse_matmul_bias=fuse_matmul_bias,
            mp_group=mp_group,
            name=name,
        )

        self._rr_column_ln = RefinedRecomputeFunction() if use_rr else None
        if self.weight.is_distributed:
            self.weight.split_axis = 1
        if has_bias and self.bias.is_distributed:
            self.bias.split_axis = 0

    def forward(self, x):
        """
        Forward pass function that computes the product of the input tensor and model parameters.

        Args:
            x (paddle.Tensor): Input tensor with shape (batch_size, seq_len, hidden_size) or (batch_size, hidden_size).
            If sequence parallel is True, the shape is (seq_len, batch_size, hidden_size).

        Returns:
            paddle.Tensor: Returns a tensor with shape (batch_size, seq_len, hidden_size) or (batch_size, hidden_size).
            If sequence parallel is True, the shape is (seq_len, batch_size, hidden_size).
        """
        # sequence parallelism is same as model parallelism
        # if sequence parallel is true, input shape is [s, b, h]
        # else input shape is [b, s, h]
        if self.is_mp:
            input_parallel = AllGatherOp.apply(x)
        else:
            input_parallel = x

        if (
            self._rr_column_ln is not None and self.training
        ):  # in eval mode, do not use refined recompute
            output = self._rr_column_ln(
                self.linear,
                x=input_parallel,
                weight=self.weight,
                bias=self.bias,
            )
        else:
            output = self.linear(
                input_parallel, self.weight, self.bias, name=self._name
            )
        return output


class RRRowSequenceParallelLinear(RowSequenceParallelLinear):
    """
    RowSequenceParallelLinear with refined recompute.
    """

    def __init__(
        self,
        in_features,
        out_features,
        weight_attr=None,
        has_bias=True,
        input_is_parallel=False,
        fuse_matmul_bias=False,
        mp_group=None,
        use_rr=False,
        name=None,
    ):
        """
        Args:
            in_features (int): The number of input features.
            out_features (int): The number of output features.
            weight_attr (ParamAttr, optional): The parameter attribute for the learnable
                weight matrix. Defaults to None. If it is None, the system will
                generate a default Attribute object.
            has_bias (bool, optional): Whether the layer uses a bias term. Defaults to True.
            input_is_parallel (bool, optional): Whether the input is parallel. Defaults to False.
            fuse_matmul_bias (bool, optional): Whether to fuse matmul and bias into one kernel. Defaults to False.
            mp_group (Group, optional): Model parallel group. Defaults to None.
            use_rr (bool, optional): Whether to use refined rr. Defaults to False.
            name (str, optional): Name of the layer. Defaults to None.
        """
        super().__init__(
            in_features=in_features,
            out_features=out_features,
            weight_attr=weight_attr,
            has_bias=has_bias,
            input_is_parallel=input_is_parallel,
            fuse_matmul_bias=fuse_matmul_bias,
            mp_group=mp_group,
            name=name,
        )

        self._rr_row_ln = RefinedRecomputeFunction() if use_rr else None

        if self.weight.is_distributed:
            self.weight.split_axis = 0

    def forward(self, x):
        """
        Forward pass function that computes the product of the input tensor and model parameters.

        Args:
            x (paddle.Tensor): Input tensor with shape (batch_size, in_features).

        Returns:
            paddle.Tensor: Returns a tensor with shape (batch_size, out_features).
        """
        input_parallel = x
        if self.is_mp:
            if self.mp_scale is not None:
                bias = self.mp_scale(self.bias, self.world_size)
            else:
                bias = None

            def linear_reduce_scatter(input, weight, bias=None, name=None):
                output = self.linear(input, weight=weight, bias=bias, name=name)
                return ReduceScatterOp.apply(output)

            if (
                self._rr_row_ln is not None and self.training
            ):  # in eval mode, do not use refined recompute
                output_ = self._rr_row_ln(
                    linear_reduce_scatter,
                    input_parallel,
                    self.weight,
                    bias=bias,
                    name=self._name,
                )
            else:
                output_ = linear_reduce_scatter(
                    input_parallel, self.weight, bias=bias, name=self._name
                )

            # if self.bias is not none, sequence parallel will use
            # register_hook to all_reduce self.bias
            if bias is None and self.bias is not None:
                output = output_ + self.bias
            else:
                output = output_
        else:
            output = self.linear(
                input_parallel, self.weight, self.bias, name=self._name
            )
        return output


class AllGatherVarlenOp(PyLayer):
    """
    A custom PyLayer that performs variable-length allgather operation.

    This operation handles tensors with different shapes across ranks by:
    1. Gathering shape information from all ranks
    2. Padding tensors to maximum size
    3. Performing allgather
    4. Reconstructing the original variable-length tensors
    """

    @staticmethod
    def forward(ctx, input):
        """Forward pass for variable-length allgather operation.

        Args:
            ctx: PyLayer context for saving state
            input (Tensor): Input tensor to be gathered (may have different sizes across ranks)

        Returns:
            Tensor: Concatenated output from all ranks with original lengths
        """
        hcg = fleet.get_hybrid_communicate_group()
        group = hcg.get_model_parallel_group()

        shape0 = paddle.to_tensor([input.shape[0]])
        shape0_all = paddle.empty(shape=[group.nranks], dtype=shape0.dtype)
        dist.stream.all_gather(shape0_all, shape0, group=group, use_calc_stream=True)
        shape0_all = shape0_all.numpy()
        max_shape0 = shape0_all.max()

        indices = []
        for idx, s in enumerate(shape0_all):
            offset = idx * max_shape0
            indices.extend(list(range(offset, offset + s)))
        indices = paddle.to_tensor(indices)

        padding = max_shape0 - input.shape[0]

        ctx.shape0 = input.shape[0]
        ctx.max_shape0 = max_shape0
        ctx.shape0_all = shape0_all
        ctx.padding = padding
        ctx.indices = indices

        if padding > 0:
            input_shape = input.shape
            input_shape[0] = padding
            padding_tensor = paddle.empty(shape=input_shape, dtype=input.dtype)
            input = paddle.concat([input, padding_tensor], axis=0)
        output = all_gather(input)
        output = paddle.gather(output, indices, axis=0)

        return output

    @staticmethod
    def backward(ctx, grad):
        """Backward pass for variable-length allgather operation.

        Args:
            ctx: PyLayer context with saved state
            grad (Tensor): Gradient flowing back through the graph

        Returns:
            Tensor: Scattered gradient with original variable lengths
        """
        input_shape = grad.shape
        input_shape[0] = ctx.max_shape0 * ctx.shape0_all.shape[0]
        output = paddle.zeros(shape=input_shape, dtype=grad.dtype)

        # grad = paddle.put_along_axis(output, ctx.indices, grad, axis=0)
        grad = paddle.scatter(output, ctx.indices, grad)
        grad = scatter(grad)

        if ctx.padding > 0:
            grad = grad[: ctx.shape0]
        return grad


def sequence_parallel_sparse_mask_labels(labels, ignore_label=-100):
    """
    Processes sparse labels in sequence parallel training by gathering non-ignored labels across all ranks.

    This function handles the case where labels may contain ignored values (typically -100) by:
    1. Distributing labels across model parallel ranks
    2. Identifying and gathering only valid (non-ignored) labels
    3. Performing a variable-length allgather operation to collect all valid labels

    Args:
        labels (paddle.Tensor): The input label tensor which may contain ignore_label values.
                              Shape should be compatible with model parallel distribution.
        ignore_label (int, optional): The value used to indicate labels that should be ignored.
                                     Defaults to -100 (common convention in NLP tasks).

    Returns:
        tuple: Contains two elements:
            - labels_all_gather (paddle.Tensor): Concatenated tensor of all non-ignored labels
                                               from all model parallel ranks.
            - tgt_index (paddle.Tensor): Indices of the non-ignored labels in the local rank's
                                        portion of the original labels tensor.

    Note:
        - This function assumes sequence parallel training is being used.
        - If a rank has no valid labels (all ignored), it will still contribute one dummy label
          (index 0) to maintain consistency in the distributed computation.
        - The returned tgt_index can be used to reconstruct the original label positions.
    """
    hcg = fleet.get_hybrid_communicate_group()
    group = hcg.get_model_parallel_group()
    labels = labels.flatten()
    labels_local = paddle.split(labels, group.nranks)[group.rank]

    tgt_index = paddle.nonzero(labels_local != ignore_label).reshape([-1])
    if tgt_index.numel() == 0:
        tgt_index = paddle.to_tensor([0])

    labels_local_gather = paddle.gather(labels_local, tgt_index, axis=0)
    labels_all_gather = AllGatherVarlenOp.apply(labels_local_gather)
    return labels_all_gather, tgt_index


async_loader = None


def get_async_loader():
    """get_async_loader"""
    global async_loader
    if not hasattr(fleet.fleet, "_hcg"):
        if async_loader is None:
            async_loader = create_async_load()
        return async_loader

    hcg = get_hcg()
    if not hasattr(hcg, "async_loader"):
        hcg.async_loader = create_async_load()
    return hcg.async_loader


def hack_offload_wait(task):
    """hack_offload_wait"""
    task.cpu_wait()


def hack_reload_wait(task):
    """hack_offload_wait"""
    task.cuda_wait()


def all_gather_group(input, group=None, axis=0):
    """Perform collective all-gather operation across a process group with axis control.

    Functional Behavior:
      - Aggregates input tensors from all processes in the specified group
      - Supports concatenation along arbitrary dimensions (axis parameter)
      - Optimizes for axis=0 via direct shape expansion to avoid concatenation overhead

    Args:
        input (Tensor):        Local tensor to be gathered (shape: [..., D, ...])
        group (ProcessGroup):  Communication group (defaults to model parallel group)
        axis (int):            Concatenation dimension (default=0)

    Returns:
        Tensor: Concatenated tensor combining inputs from all processes:
                - When axis=0: shape [D*N, ...] (N = group size)
                - Otherwise:   shape [..., D*N, ...] along specified axis
    """
    if group is None:
        hcg = fleet.get_hybrid_communicate_group()
        group = hcg.get_model_parallel_group()
    parallelism = group.nranks
    if parallelism == 1:
        return input.clone()
    output_shape = input.shape
    if axis == 0:
        output_shape[axis] = output_shape[axis] * parallelism
        output = paddle.empty(shape=output_shape, dtype=input.dtype)
        dist.stream.all_gather(output, input, group=group, use_calc_stream=True)
        return output
    outputs = [
        paddle.empty(output_shape, dtype=input.dtype) for _ in range(parallelism)
    ]
    dist.stream.all_gather(outputs, input, group=group, use_calc_stream=True)
    output = paddle.concat(outputs, axis=axis)
    return output


def reduce_scatter_group(input, group=None):
    """Perform reduce-scatter collective operation across a process group.

    Functional Behavior:
      - Aggregates (sums) input tensors across all processes in the group
      - Scatters the reduced result equally to all participants
      - Operates along the first dimension (axis=0) of the input tensor

    Args:
        input (Tensor):        Local tensor to reduce (shape: [N*K, ...] where N=group_size)
        group (ProcessGroup): Communication group (defaults to model parallel group)

    Returns:
        Tensor: Scattered portion of reduced tensor with shape [K, ...]
    """
    if group is None:
        hcg = fleet.get_hybrid_communicate_group()
        group = hcg.get_model_parallel_group()
    parallelism = group.nranks
    if parallelism == 1:
        return input.clone()
    output_shape = input.shape
    assert (
        input.shape[0] % parallelism == 0
    ), f"Input sequence length {input.shape[0]} can't be divided exactly by sequence parallelism {parallelism}"
    output_shape[0] = output_shape[0] // parallelism
    output = paddle.empty(shape=output_shape, dtype=input.dtype)
    dist.stream.reduce_scatter(
        output, input, op=dist.ReduceOp.SUM, group=group, use_calc_stream=True
    )
    return output



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Final, List, Optional, Tuple, Union

from pydantic import BaseModel, Field, field_validator, model_validator
from typing_extensions import Annotated, Literal

from ..infra.models import DataInfo, PrimaryOperations
from .shared import ocr

__all__ = [
    "INFER_ENDPOINT",
    "InferRequest",
    "LayoutParsingResult",
    "InferResult",
    "RESTRUCTURE_PAGES_ENDPOINT",
    "RestructurePagesRequest",
    "RestructurePagesResult",
    "PRIMARY_OPERATIONS",
    "MarkdownData",
    "Page",
]

INFER_ENDPOINT: Final[str] = "/layout-parsing"
RESTRUCTURE_PAGES_ENDPOINT: Final[str] = "/restructure-pages"


class InferRequest(ocr.BaseInferRequest):
    useDocOrientationClassify: Optional[bool] = None
    useDocUnwarping: Optional[bool] = None
    useLayoutDetection: Optional[bool] = None
    useChartRecognition: Optional[bool] = None
    useSealRecognition: Optional[bool] = None
    useOcrForImageBlock: Optional[bool] = None
    layoutThreshold: Optional[Union[float, dict]] = None
    layoutNms: Optional[bool] = None
    layoutUnclipRatio: Optional[Union[float, Tuple[float, float], dict]] = None
    layoutMergeBboxesMode: Optional[Union[str, dict]] = None
    layoutShapeMode: Literal["rect", "quad", "poly", "auto"] = "auto"
    promptLabel: Optional[str] = None
    formatBlockContent: Optional[bool] = None
    repetitionPenalty: Optional[float] = None
    temperature: Optional[float] = None
    topP: Optional[float] = None
    minPixels: Optional[Annotated[int, Field(gt=0)]] = None
    maxPixels: Optional[Annotated[int, Field(gt=0)]] = None
    maxNewTokens: Optional[Annotated[int, Field(gt=0)]] = None
    mergeLayoutBlocks: Optional[bool] = None
    markdownIgnoreLabels: Optional[List[str]] = None
    vlmExtraArgs: Optional[dict] = None
    prettifyMarkdown: bool = True
    showFormulaNumber: bool = False
    restructurePages: bool = False
    mergeTables: bool = True
    relevelTitles: bool = True
    visualize: Optional[bool] = None
    logId: Optional[str] = None

    @field_validator("topP")
    @classmethod
    def validate_top_p(cls, v):
        if v is not None and not (0 < v <= 1):
            raise ValueError(f"`topP` must be > 0 and ≤ 1; got: {v}")
        return v

    @field_validator("temperature")
    @classmethod
    def validate_temperature(cls, v):
        if v is not None and v < 0:
            raise ValueError(f"`temperature` must be ≥ 0; got: {v}")
        return v

    @field_validator("repetitionPenalty")
    @classmethod
    def validate_repetition_penalty(cls, v):
        if v is not None and v <= 0:
            raise ValueError(f"`repetitionPenalty` must be > 0; got: {v}")
        return v

    @field_validator("promptLabel")
    @classmethod
    def validate_prompt_label(cls, v):
        _ALLOWED_PROMPT_LABELS = (
            "ocr",
            "formula",
            "table",
            "chart",
            "seal",
            "spotting",
        )
        if v is not None and v not in _ALLOWED_PROMPT_LABELS:
            valid_values = ", ".join(_ALLOWED_PROMPT_LABELS)
            raise ValueError(f"`promptLabel` must be one of: {valid_values}; got: {v}")
        return v

    @field_validator("layoutMergeBboxesMode")
    @classmethod
    def validate_merge_bboxes_mode(cls, v):
        _ALLOWED_MERGE_BBOXES_MODES = ("large", "small", "union")

        if v is None:
            return v

        if isinstance(v, str):
            if v not in _ALLOWED_MERGE_BBOXES_MODES:
                raise ValueError(
                    f"`layoutMergeBboxesMode` must be one of: {', '.join(_ALLOWED_MERGE_BBOXES_MODES)}; got: {v}"
                )
        elif isinstance(v, dict):
            for key, value in v.items():
                if not isinstance(value, str):
                    raise ValueError(
                        f"`layoutMergeBboxesMode` dictionary values must be strings; got: {type(value).__name__}"
                    )
                if value not in _ALLOWED_MERGE_BBOXES_MODES:
                    raise ValueError(
                        f"`layoutMergeBboxesMode` dictionary value must be one of: {', '.join(_ALLOWED_MERGE_BBOXES_MODES)}; got: {value}"
                    )
        else:
            raise ValueError("`layoutMergeBboxesMode` must be a string or dictionary")

        return v

    @field_validator("layoutUnclipRatio")
    @classmethod
    def validate_unclip_ratio(cls, v):
        if v is None:
            return v

        def _validate_ratio_value(value, context=""):
            if isinstance(value, (int, float)):
                if value <= 0:
                    raise ValueError(
                        f"`layoutUnclipRatio`{context} must be > 0; got: {value}"
                    )
            elif isinstance(value, list):
                if len(value) != 2:
                    raise ValueError(
                        f"`layoutUnclipRatio`{context} must be two numbers; got: {len(value)} values"
                    )
                for i, item in enumerate(value):
                    if not isinstance(item, (int, float)):
                        raise ValueError(
                            f"`layoutUnclipRatio`{context} values must be numbers; got: {type(item).__name__} at position {i}"
                        )
                    if item <= 0:
                        raise ValueError(
                            f"`layoutUnclipRatio`{context} values must be > 0; got: {item} at position {i}"
                        )
            else:
                raise ValueError(
                    f"`layoutUnclipRatio`{context} must be a number or two numbers; got: {type(value).__name__}"
                )

        if isinstance(v, dict):
            for key, value in v.items():
                _validate_ratio_value(value, f" value for key '{key}'")
        else:
            _validate_ratio_value(v)

        return v

    @field_validator("layoutThreshold")
    @classmethod
    def validate_threshold(cls, v):
        if v is None:
            return v

        def _validate_threshold_value(value, context=""):
            if not isinstance(value, (int, float)):
                raise ValueError(
                    f"`layoutThreshold`{context} must be a number; got: {type(value).__name__}"
                )
            if value < 0 or value > 1:
                raise ValueError(
                    f"`layoutThreshold`{context} must be between 0 and 1 inclusive; got: {value}"
                )

        if isinstance(v, dict):
            for key, value in v.items():
                _validate_threshold_value(value, f" value for key '{key}'")
        else:
            _validate_threshold_value(v)

        return v

    @model_validator(mode="after")
    def validate_pixel_range(self):
        if self.minPixels is not None and self.maxPixels is not None:
            if self.minPixels > self.maxPixels:
                raise ValueError(
                    f"`minPixels` ({self.minPixels}) cannot be greater than `maxPixels` ({self.maxPixels})"
                )
        return self


class MarkdownData(BaseModel):
    text: str
    images: Optional[Dict[str, str]] = None


class LayoutParsingResult(BaseModel):
    prunedResult: dict
    markdown: MarkdownData
    outputImages: Optional[Dict[str, str]] = None
    inputImage: Optional[str] = None


class InferResult(BaseModel):
    layoutParsingResults: List[LayoutParsingResult]
    dataInfo: DataInfo


class Page(BaseModel):
    prunedResult: dict
    markdownImages: Optional[Dict[str, str]] = None


class RestructurePagesRequest(BaseModel):
    pages: List[Page]
    mergeTables: bool = True
    relevelTitles: bool = True
    concatenatePages: bool = False
    prettifyMarkdown: bool = True
    showFormulaNumber: bool = False
    logId: Optional[str] = None


class RestructurePagesResult(BaseModel):
    layoutParsingResults: List[LayoutParsingResult]


PRIMARY_OPERATIONS: Final[PrimaryOperations] = {
    "infer": (INFER_ENDPOINT, InferRequest, InferResult),
    "restructure-pages": (
        RESTRUCTURE_PAGES_ENDPOINT,
        RestructurePagesRequest,
        RestructurePagesResult,
    ),
}



# Copyright (c) 2025 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
npu fusion operators.

"""
import paddle
import paddle.distributed as dist
import paddle.nn.functional as F


def npu_combining(x, combine_weights, scatter_index, hard_gate=False):
    """
    Args:
        x: Tensor[seq, dim]
        combine_weights: [seq, k]
        scatter_index:  ** [seq, k] **
    Returns:
        y: Tensor[s, dim]
    """
    x_gatherd = F.embedding(scatter_index, x)  # [s,k,dim]
    if hard_gate:
        return x_gatherd.squeeze(-2)
    y = (combine_weights.unsqueeze(-1) * x_gatherd).sum(1)
    return y


def npu_cal_aux_loss_func(
    gate_prob,
    dispatch_mask,
    tokens_mask,
    dispatch_tokens_mask,
    num_experts,
    use_group,
    moe_k,
    global_aux_loss=False,
    rank=None,
    group=None,
    clip_min=1e-6,
):
    """cal_aux_loss_func"""
    if tokens_mask is not None and tokens_mask.dtype != gate_prob.dtype:
        tokens_mask = tokens_mask.astype(gate_prob.dtype)

    scale = None
    if dispatch_tokens_mask is not None:
        seqlen_float = dispatch_tokens_mask.astype(gate_prob.dtype).sum()
        if (
            tokens_mask is not None
            and gate_prob.shape[0] != dispatch_tokens_mask.shape[0]
        ):
            scale = seqlen_float / paddle.clip(tokens_mask.sum(), min=1e-6)
    elif tokens_mask is not None:
        seqlen_float = tokens_mask.sum()
    else:
        seqlen_float = gate_prob.numel().astype(gate_prob.dtype) / num_experts
    seqlen_float = paddle.clip(seqlen_float, min=1e-6)
    if len(dispatch_mask.shape) == 2:
        dispatch_mask = dispatch_mask.sum(0)
    ce = dispatch_mask.astype(gate_prob.dtype).detach() / seqlen_float
    me = paddle.sum(gate_prob, axis=0) / seqlen_float

    if global_aux_loss:
        me_list, ce_list = [], []
        dist.all_gather(me_list, me, group=group)
        dist.all_gather(ce_list, ce, group=group)
        me_list[rank] = me
        ce_list[rank] = ce
        me = paddle.stack(me_list).mean(0)
        ce = paddle.stack(ce_list).mean(0)

    l_aux = paddle.sum(me * ce) * num_experts
    if use_group:
        l_aux = l_aux / moe_k
    if scale is not None:
        l_aux = l_aux + (scale - 1) * l_aux.detach()
    return l_aux, None, None



# Copyright (c) 2025 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pycocotools.mask as mask_util
from paddlex_hps_server import BaseTritonPythonModel, schemas, utils


def _rle(mask: np.ndarray) -> str:
    rle_res = mask_util.encode(np.asarray(mask[..., None], order="F", dtype="uint8"))[0]
    return rle_res["counts"].decode("utf-8")


class TritonPythonModel(BaseTritonPythonModel):
    def get_input_model_type(self):
        return schemas.open_vocabulary_segmentation.InferRequest

    def get_result_model_type(self):
        return schemas.open_vocabulary_segmentation.InferResult

    def run(self, input, log_id):
        file_bytes = utils.get_raw_bytes(input.image)
        image = utils.image_bytes_to_array(file_bytes)
        visualize_enabled = (
            input.visualize
            if input.visualize is not None
            else self.app_config.visualize
        )

        result = list(
            self.pipeline.predict(
                image,
                prompt=input.prompt,
                prompt_type=input.promptType,
            )
        )[0]

        rle_masks = [
            dict(rleResult=_rle(mask), size=mask.shape) for mask in result["masks"]
        ]
        mask_infos = result["mask_infos"]
        if visualize_enabled:
            output_image_base64 = utils.base64_encode(
                utils.image_to_bytes(result.img["res"])
            )
        else:
            output_image_base64 = None

        return schemas.open_vocabulary_segmentation.InferResult(
            masks=rle_masks, maskInfos=mask_infos, image=output_image_base64
        )



# Copyright (c) 2025 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ....utils import logging
from ....utils.deps import (
    get_dep_version,
    is_genai_engine_plugin_available,
    require_genai_engine_plugin,
)
from ..configs.utils import (
    backend_config_to_args,
    set_config_defaults,
    update_backend_config,
)
from ..models import ALL_MODEL_INFO, get_model_components, is_integrated_model_available


def register_models():
    from vllm import ModelRegistry

    vllm_version = get_dep_version("vllm")

    if is_genai_engine_plugin_available("vllm-server"):
        for model_name in ALL_MODEL_INFO:
            if (
                not is_integrated_model_available(model_name, "vllm")
                and model_name not in ModelRegistry.get_supported_archs()
            ):
                net_cls, _ = get_model_components(model_name, "vllm")
                ModelRegistry.register_model(net_cls.__name__, net_cls)


def run_vllm_server(host, port, model_name, model_dir, config, chat_template_path):
    require_genai_engine_plugin("vllm-server")

    import uvloop
    from vllm.entrypoints.openai.api_server import (
        FlexibleArgumentParser,
        cli_env_setup,
        make_arg_parser,
        run_server,
        validate_parsed_serve_args,
    )

    cli_env_setup()
    parser = FlexibleArgumentParser()
    parser = make_arg_parser(parser)

    set_config_defaults(config, {"served-model-name": model_name})

    if chat_template_path:
        set_config_defaults(config, {"chat-template": str(chat_template_path)})

    update_backend_config(
        config,
        {
            "model": model_dir,
            "host": host,
            "port": port,
        },
    )

    import torch

    if torch.version.hip is not None and torch.version.cuda is None:
        # For DCU
        if "api-server-count" in config:
            logging.warning(
                "Key 'api-server-count' will be popped as it is not supported"
            )
            config.pop("api-server-count")

    args = backend_config_to_args(config)
    args = parser.parse_args(args)
    validate_parsed_serve_args(args)

    uvloop.run(run_server(args))



# Copyright (c) 2026 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

PIPELINE_APP_ROUTER = {
    "PaddleOCR-VL-1.5": "PaddleOCR-VL",
}


def pipeline_name_to_mod_name(pipeline_name: str) -> str:
    if not pipeline_name:
        raise ValueError("Empty pipeline name")
    if pipeline_name in PIPELINE_APP_ROUTER:
        pipeline_name = PIPELINE_APP_ROUTER[pipeline_name]
    mod_name = pipeline_name.lower().replace("-", "_").replace(".", "")
    if mod_name[0].isdigit():
        return "m_" + mod_name
    return mod_name



# Copyright (c) 2025 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Any, Dict, List

from paddlex_hps_server import BaseTritonPythonModel, protocol, schemas, utils


class TritonPythonModel(BaseTritonPythonModel):
    def get_input_model_type(self):
        return schemas.multilingual_speech_recognition.InferRequest

    def get_result_model_type(self):
        return schemas.multilingual_speech_recognition.InferResult

    def run(self, input, log_id):
        file_bytes = utils.get_raw_bytes(input.audio)
        ext = utils.infer_file_ext(input.audio)
        if ext is None:
            return protocol.create_aistudio_output_without_result(
                422,
                "File extension cannot be inferred",
                log_id=log_id,
            )
        audio_path = utils.write_to_temp_file(
            file_bytes,
            suffix=ext,
        )

        try:
            result = list(self.pipeline(audio_path))[0]
        finally:
            os.unlink(audio_path)

        segments: List[Dict[str, Any]] = []
        for item in result["result"]["segments"]:
            segment = dict(
                id=item["id"],
                seek=item["seek"],
                start=item["start"],
                end=item["end"],
                text=item["text"],
                tokens=item["tokens"],
                temperature=item["temperature"],
                avgLogProb=item["avg_logprob"],
                compressionRatio=item["compression_ratio"],
                noSpeechProb=item["no_speech_prob"],
            )
            segments.append(segment)

        return schemas.multilingual_speech_recognition.InferResult(
            text=result["result"]["text"],
            segments=segments,
            language=result["result"]["language"],
        )



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import logging

from ..model import BaseUltraInferModel
from ..runtime import Runtime, RuntimeOption

_logger = logging.getLogger(__name__)


class PyOnlyUltraInferModel(BaseUltraInferModel):
    def __init__(self, option):
        super().__init__()
        if option is None:
            self._option = RuntimeOption()
        else:
            self._option = option
        self._update_option()
        self._runtime = Runtime(self._option)
        _logger.debug("Python-only model initialized")

    def num_inputs_of_runtime(self):
        return self._runtime.num_inputs()

    def num_outputs_of_runtime(self):
        return self._runtime.num_outputs()

    def get_profile_time(self):
        return self._runtime.get_profile_time()

    def _update_option(self):
        pass


class PyOnlyProcessor(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def __call__(self, data):
        raise NotImplementedError


class PyOnlyProcessorChain(object):
    def __init__(self, processors):
        super().__init__()
        self._processors = processors

    def __call__(self, data):
        for processor in self._processors:
            data = processor(data)
        return data



# Copyright (c) 2025 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Acknowledgement: The logic for table merging in this function is adapted from MinerU.


def full_to_half(text: str) -> str:
    result = []
    for char in text:
        code = ord(char)
        if 0xFF01 <= code <= 0xFF5E:
            result.append(chr(code - 0xFEE0))
        else:
            result.append(char)
    return "".join(result)


def calculate_table_total_columns(soup):
    """
    calculate total columns including colspan and rowspan, accounting for merged cells
    """
    rows = soup.find_all("tr")
    if not rows:
        return 0
    max_cols = 0
    occupied = {}
    for row_idx, row in enumerate(rows):
        col_idx = 0
        cells = row.find_all(["td", "th"])
        if row_idx not in occupied:
            occupied[row_idx] = {}
        for cell in cells:
            while col_idx in occupied[row_idx]:
                col_idx += 1
            colspan = int(cell.get("colspan", 1))
            rowspan = int(cell.get("rowspan", 1))
            for r in range(row_idx, row_idx + rowspan):
                if r not in occupied:
                    occupied[r] = {}
                for c in range(col_idx, col_idx + colspan):
                    occupied[r][c] = True
            col_idx += colspan
            max_cols = max(max_cols, col_idx)
    return max_cols


def calculate_row_columns(row):
    """
    Calculate the actual number of columns in a single row
    """
    return sum(int(cell.get("colspan", 1)) for cell in row.find_all(["td", "th"]))


def calculate_visual_columns(row):
    """
    Calculate the visual number of columns in a single row, excluding colspan (merged cells count as one)
    """
    return len(row.find_all(["td", "th"]))


def detect_table_headers(soup1, soup2, max_header_rows=5):
    """
    Determine how many identical rows exist at the beginning of two tables
    """
    rows1 = soup1.find_all("tr")
    rows2 = soup2.find_all("tr")
    # Check only the minimum number of rows
    min_rows = min(len(rows1), len(rows2), max_header_rows)
    header_rows = 0
    headers_match = True
    for i in range(min_rows):
        cells1 = rows1[i].find_all(["td", "th"])
        cells2 = rows2[i].find_all(["td", "th"])
        if len(cells1) != len(cells2):
            headers_match = header_rows > 0
            break
        # If column counts match, check if content is identical
        match = True
        for c1, c2 in zip(cells1, cells2):
            text1 = "".join(full_to_half(c1.get_text()).split())
            text2 = "".join(full_to_half(c2.get_text()).split())
            if text1 != text2 or int(c1.get("colspan", 1)) != int(c2.get("colspan", 1)):
                match = False
                break
        # Complete match, increment matched row count. Otherwise, stop matching.
        if match:
            header_rows += 1
        else:
            headers_match = header_rows > 0
            break
    if header_rows == 0:
        headers_match = False
    return header_rows, headers_match


def check_rows_match(soup1, soup2):
    rows1 = soup1.find_all("tr")
    rows2 = soup2.find_all("tr")
    if not rows1 or not rows2:
        return False
    last_row = rows1[-1]
    header_count, _ = detect_table_headers(soup1, soup2)
    first_data_row = rows2[header_count] if len(rows2) > header_count else None
    if not first_data_row:
        return False
    last_cols = calculate_row_columns(last_row)
    first_cols = calculate_row_columns(first_data_row)
    last_visual = calculate_visual_columns(last_row)
    first_visual = calculate_visual_columns(first_data_row)
    return last_cols == first_cols or last_visual == first_visual


def is_skippable(block, allowed_labels):

    continue_keywords = ["continue", "continued", "cont'd", "续", "cont‘d", "續"]

    if block.label in allowed_labels:
        return True

    b_text = str(getattr(block, "text", "") or "").lower()
    b_fig_title = str(getattr(block, "figure_title", "") or "").lower()
    b_doc_title = str(getattr(block, "doc_title", "") or "").lower()
    b_para_title = str(getattr(block, "paragraph_title", "") or "").lower()

    full_content = f"{b_text} {b_fig_title} {b_doc_title} {b_para_title}"

    if any(kw in full_content for kw in continue_keywords):
        return True

    return False


def can_merge_tables(prev_page, prev_block, curr_page, curr_block):

    from bs4 import BeautifulSoup

    x0, y0, x1, y1 = prev_block.bbox
    prev_width = x1 - x0
    x2, y2, x3, y4 = curr_block.bbox
    curr_width = x3 - x2
    if curr_width == 0 or prev_width == 0:
        return False, None, None
    if abs(curr_width - prev_width) / min(curr_width, prev_width) >= 0.1:
        return False, None, None

    prev_index = prev_page.index(prev_block)
    allowed_follow = all(
        b.label
        in ["footer", "vision_footnote", "number", "footnote", "footer_image", "seal"]
        for b in prev_page[prev_index + 1 :]
    )
    if not allowed_follow:
        return False, None, None

    curr_index = curr_page.index(curr_block)
    curr_allowed_labels = ["header", "header_image", "number", "seal"]

    allowed_before = all(
        is_skippable(b, curr_allowed_labels) for b in curr_page[:curr_index]
    )
    if not allowed_before:
        return False, None, None

    html_prev = prev_block.content
    html_curr = curr_block.content
    if not html_prev or not html_curr:
        return False, None, None
    soup_prev = BeautifulSoup(html_prev, "html.parser")
    soup_curr = BeautifulSoup(html_curr, "html.parser")

    total_cols_prev = calculate_table_total_columns(soup_prev)
    total_cols_curr = calculate_table_total_columns(soup_curr)
    tables_match = total_cols_prev == total_cols_curr
    rows_match = check_rows_match(soup_prev, soup_curr)

    return (tables_match or rows_match), soup_prev, soup_curr


def perform_table_merge(soup_prev, soup_curr):
    header_count, _ = detect_table_headers(soup_prev, soup_curr)
    rows_prev = soup_prev.find_all("tr")
    rows_curr = soup_curr.find_all("tr")
    for row in rows_curr[header_count:]:
        row.extract()
        rows_prev[-1].parent.append(row)
    return str(soup_prev)


def merge_tables_across_pages(pages):
    for i in range(len(pages) - 1, 0, -1):
        page_curr = pages[i]
        page_prev = pages[i - 1]

        for block in page_curr:
            if block.label == "table":
                curr_block = block
                break
        else:
            curr_block = None

        for block in reversed(page_prev):
            if block.label == "table":
                prev_block = block
                break
        else:
            prev_block = None

        # both curr_block and prev_block should not be None
        if curr_block and prev_block:
            can_merge, soup_prev, soup_curr = can_merge_tables(
                page_prev, prev_block, page_curr, curr_block
            )
        else:
            can_merge = False

        if can_merge:
            merged_html = perform_table_merge(soup_prev, soup_curr)
            prev_block.content = merged_html
            prev_block_global_id = prev_block.global_block_id
            curr_block.content = ""
            curr_block.global_group_id = prev_block_global_id
    all_blocks = [block for page in pages for block in page]
    for page in pages:
        for block in page:
            if block.global_block_id != block.global_group_id:
                block.global_group_id = all_blocks[
                    block.global_group_id
                ].global_group_id
    return pages



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from paddlex import create_pipeline

pipeline = create_pipeline(pipeline="PP-ChatOCRv4-doc")

img_path = "./test_samples/研报2_11.jpg"
key_list = ["三位一体养老生态系统包含哪些"]

# img_path = "./test_samples/财报1.pdf"
# key_list = ['公司全称是什么']

visual_predict_res = pipeline.visual_predict(
    img_path,
    use_doc_orientation_classify=False,
    use_doc_unwarping=False,
    use_common_ocr=True,
    use_seal_recognition=True,
    use_table_recognition=True,
)

visual_info_list = []
for res in visual_predict_res:
    visual_info_list.append(res["visual_info"])
    layout_parsing_result = res["layout_parsing_result"]
    print(layout_parsing_result)
    layout_parsing_result.print()
    layout_parsing_result.save_to_img("./output")
    layout_parsing_result.save_to_json("./output")
    layout_parsing_result.save_to_xlsx("./output")
    layout_parsing_result.save_to_html("./output")

pipeline.save_visual_info_list(
    visual_info_list, "./res_visual_info/tmp_visual_info.json"
)

visual_info_list = pipeline.load_visual_info_list(
    "./res_visual_info/tmp_visual_info.json"
)

vector_info = pipeline.build_vector(visual_info_list, flag_save_bytes_vector=True)

pipeline.save_vector(vector_info, "./res_visual_info/tmp_vector_info.json")

vector_info = pipeline.load_vector("./res_visual_info/tmp_vector_info.json")

mllm_predict_res = pipeline.mllm_pred(input=img_path, key_list=key_list)
mllm_predict_info = mllm_predict_res["mllm_res"]

chat_result = pipeline.chat(
    key_list,
    visual_info_list,
    vector_info=vector_info,
    mllm_predict_info=mllm_predict_info,
)

print(chat_result)



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import logging

from ... import ModelFormat, RuntimeOption, UltraInferModel
from ... import c_lib_wrap as C


class SchemaLanguage(object):
    ZH = 0
    EN = 1


class SchemaNode(object):
    def __init__(self, name, children=[]):
        schema_node_children = []
        if isinstance(children, str):
            children = [children]
        for child in children:
            if isinstance(child, str):
                schema_node_children += [C.text.SchemaNode(child, [])]
            elif isinstance(child, dict):
                for key, val in child.items():
                    schema_node_child = SchemaNode(key, val)
                    schema_node_children += [schema_node_child._schema_node]
            else:
                assert "The type of child of SchemaNode should be str or dict."
        self._schema_node = C.text.SchemaNode(name, schema_node_children)
        self._schema_node_children = schema_node_children


class UIEModel(UltraInferModel):
    def __init__(
        self,
        model_file,
        params_file,
        vocab_file,
        position_prob=0.5,
        max_length=128,
        schema=[],
        batch_size=64,
        runtime_option=RuntimeOption(),
        model_format=ModelFormat.PADDLE,
        schema_language=SchemaLanguage.ZH,
    ):
        if isinstance(schema, list):
            schema = SchemaNode("", schema)._schema_node_children
        elif isinstance(schema, dict):
            schema_tmp = []
            for key, val in schema.items():
                schema_tmp += [SchemaNode(key, val)._schema_node]
            schema = schema_tmp
        else:
            assert "The type of schema should be list or dict."
        schema_language = C.text.SchemaLanguage(schema_language)
        self._model = C.text.UIEModel(
            model_file,
            params_file,
            vocab_file,
            position_prob,
            max_length,
            schema,
            batch_size,
            runtime_option._option,
            model_format,
            schema_language,
        )
        assert self.initialized, "UIEModel initialize failed."

    def set_schema(self, schema):
        if isinstance(schema, list):
            schema = SchemaNode("", schema)._schema_node_children
        elif isinstance(schema, dict):
            schema_tmp = []
            for key, val in schema.items():
                schema_tmp += [SchemaNode(key, val)._schema_node]
            schema = schema_tmp
        self._model.set_schema(schema)

    def predict(self, texts, return_dict=False):
        results = self._model.predict(texts)
        if not return_dict:
            return results
        new_results = []
        for result in results:
            uie_result = dict()
            for key, uie_results in result.items():
                uie_result[key] = list()
                for uie_res in uie_results:
                    uie_result[key].append(uie_res.get_dict())
            new_results += [uie_result]
        return new_results



# Copyright (c) 2025 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import paddle

from ....utils.benchmark import add_inference_operations, benchmark
from ...common.transformers.transformers import PretrainedConfig, PretrainedModel
from .pp_ocrv5_rec_modules.rec_lcnetv3 import PPLCNetV3
from .pp_ocrv5_rec_modules.rec_multi_head import MultiHead
from .pp_ocrv5_rec_modules.rec_pphgnetv2 import PPHGNetV2

__all__ = ["PPOCRV5Rec"]


class PPOCRV5RecConfig(PretrainedConfig):
    def __init__(
        self,
        backbone,
        MultiHead,
    ):
        self.backbone_name = backbone["name"]
        if self.backbone_name == "PPLCNetV3":
            self.net_config = backbone["net_config"]
            self.scale = backbone["scale"]
            self.conv_kxk_num = backbone["conv_kxk_num"]
            self.lr_mult_list = backbone["lr_mult_list"]
            self.lab_lr = backbone["lab_lr"]
        elif self.backbone_name == "PPHGNetV2":
            self.text_rec = backbone["text_rec"]
            self.stem_channels = backbone["stem_channels"]
            self.stage_config = backbone["stage_config"]
            self.det = backbone["det"]
            self.use_lab = backbone["use_lab"]
            self.use_last_conv = backbone["use_last_conv"]
            self.class_expand = backbone["class_expand"]
            self.dropout_prob = backbone["dropout_prob"]
            self.class_num = backbone["class_num"]
            self.lr_mult_list = backbone["lr_mult_list"]
            self.out_indices = backbone["out_indices"]
        else:
            raise RuntimeError(
                f"There is no dynamic graph implementation for backbone {backbone['name']}."
            )
        self.head_list = MultiHead["head_list"]
        self.decode_list = MultiHead["decode_list"]
        self.tensor_parallel_degree = 1


class PPOCRV5Rec(PretrainedModel):

    config_class = PPOCRV5RecConfig

    def __init__(self, config: PPOCRV5RecConfig):
        super().__init__(config)
        if self.config.backbone_name == "PPLCNetV3":
            self.backbone = PPLCNetV3(
                scale=self.config.scale,
                net_config=self.config.net_config,
                conv_kxk_num=self.config.conv_kxk_num,
                lr_mult_list=self.config.lr_mult_list,
                lab_lr=self.config.lab_lr,
            )
        elif self.config.backbone_name == "PPHGNetV2":
            self.backbone = PPHGNetV2(
                stage_config=self.config.stage_config,
                stem_channels=self.config.stem_channels,
                text_rec=self.config.text_rec,
                det=self.config.det,
                use_lab=self.config.use_lab,
                use_last_conv=self.config.use_last_conv,
                class_expand=self.config.class_expand,
                dropout_prob=self.config.dropout_prob,
                class_num=self.config.class_num,
                lr_mult_list=self.config.lr_mult_list,
                out_indices=self.config.out_indices,
            )
        self.head = MultiHead(
            in_channels=self.backbone.out_channels,
            out_channels_list=self.config.decode_list,
            head_list=self.config.head_list,
        )

    add_inference_operations("pp_ocrv5_rec_forward")

    @benchmark.timeit_with_options(name="pp_ocrv5_rec_forward")
    def forward(self, x):
        x = paddle.to_tensor(x[0])
        x = self.backbone(x)
        x = self.head(x)
        return [x.cpu().numpy()]

    def get_transpose_weight_keys(self):
        transpose_keys = ["fc", "out_proj", "attn.qkv"]
        need_to_transpose = []
        all_weight_keys = []
        for name, param in self.head.named_parameters():
            all_weight_keys.append("head." + name)
        for i in range(len(all_weight_keys)):
            for j in range(len(transpose_keys)):
                if (transpose_keys[j] in all_weight_keys[i]) and (
                    "bias" not in all_weight_keys[i]
                ):
                    need_to_transpose.append(all_weight_keys[i])
        if self.config.backbone_name == "PPHGNetV2":
            need_to_transpose.append("backbone.fc.weight")
        return need_to_transpose

    def get_hf_state_dict(self, *args, **kwargs):

        model_state_dict = self.state_dict(*args, **kwargs)

        hf_state_dict = {}
        for old_key, value in model_state_dict.items():
            if "_mean" in old_key:
                new_key = old_key.replace("_mean", "running_mean")
            elif "_variance" in old_key:
                new_key = old_key.replace("_variance", "running_var")
            else:
                new_key = old_key
            hf_state_dict[new_key] = value

        return hf_state_dict

    def set_hf_state_dict(self, state_dict, *args, **kwargs):

        key_mapping = {}
        for old_key in list(state_dict.keys()):
            if "running_mean" in old_key:
                key_mapping[old_key] = old_key.replace("running_mean", "_mean")
            elif "running_var" in old_key:
                key_mapping[old_key] = old_key.replace("running_var", "_variance")

        for old_key, new_key in key_mapping.items():
            state_dict[new_key] = state_dict.pop(old_key)

        return self.set_state_dict(state_dict, *args, **kwargs)



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import os.path as osp
import shutil
import requests
import time
import zipfile
import tarfile
import hashlib
import tqdm
import logging

from .utils.hub_model_server import model_server
from .utils import hub_env as hubenv

DOWNLOAD_RETRY_LIMIT = 3


def md5check(fullname, md5sum=None):
    if md5sum is None:
        return True

    logging.info("File {} md5 checking...".format(fullname))
    md5 = hashlib.md5()
    with open(fullname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            md5.update(chunk)
    calc_md5sum = md5.hexdigest()

    if calc_md5sum != md5sum:
        logging.info(
            "File {} md5 check failed, {}(calc) != "
            "{}(base)".format(fullname, calc_md5sum, md5sum)
        )
        return False
    return True


def move_and_merge_tree(src, dst):
    """
    Move src directory to dst, if dst is already exists,
    merge src to dst
    """
    if not osp.exists(dst):
        shutil.move(src, dst)
    else:
        if not osp.isdir(src):
            shutil.move(src, dst)
            return
        for fp in os.listdir(src):
            src_fp = osp.join(src, fp)
            dst_fp = osp.join(dst, fp)
            if osp.isdir(src_fp):
                if osp.isdir(dst_fp):
                    move_and_merge_tree(src_fp, dst_fp)
                else:
                    shutil.move(src_fp, dst_fp)
            elif osp.isfile(src_fp) and not osp.isfile(dst_fp):
                shutil.move(src_fp, dst_fp)


def download(url, path, rename=None, md5sum=None, show_progress=False):
    """
    Download from url, save to path.
    url (str): download url
    path (str): download to given path
    """
    if not osp.exists(path):
        os.makedirs(path)

    fname = osp.split(url)[-1]
    fullname = osp.join(path, fname)
    if rename is not None:
        fullname = osp.join(path, rename)
    retry_cnt = 0
    while not (osp.exists(fullname) and md5check(fullname, md5sum)):
        if retry_cnt < DOWNLOAD_RETRY_LIMIT:
            retry_cnt += 1
        else:
            logging.debug("{} download failed.".format(fname))
            raise RuntimeError(
                "Download from {} failed. " "Retry limit reached".format(url)
            )

        logging.info("Downloading {} from {}".format(fname, url))

        req = requests.get(url, stream=True)
        if req.status_code != 200:
            raise RuntimeError(
                "Downloading from {} failed with code "
                "{}!".format(url, req.status_code)
            )

        # For protecting download interrupted, download to
        # tmp_fullname firstly, move tmp_fullname to fullname
        # after download finished
        tmp_fullname = fullname + "_tmp"
        total_size = req.headers.get("content-length")
        with open(tmp_fullname, "wb") as f:
            if total_size and show_progress:
                for chunk in tqdm.tqdm(
                    req.iter_content(chunk_size=1024),
                    total=(int(total_size) + 1023) // 1024,
                    unit="KB",
                ):
                    f.write(chunk)
            else:
                for chunk in req.iter_content(chunk_size=1024):
                    if chunk:
                        f.write(chunk)
        shutil.move(tmp_fullname, fullname)
        logging.debug("{} download completed.".format(fname))

    return fullname


def decompress(fname):
    """
    Decompress for zip and tar file
    """
    logging.info("Decompressing {}...".format(fname))

    # For protecting decompressing interrupted,
    # decompress to fpath_tmp directory firstly, if decompress
    # succeeded, move decompress files to fpath and delete
    # fpath_tmp and remove download compress file.
    fpath = osp.split(fname)[0]
    fpath_tmp = osp.join(fpath, "tmp")
    if osp.isdir(fpath_tmp):
        shutil.rmtree(fpath_tmp)
        os.makedirs(fpath_tmp)

    if fname.find(".tar") >= 0 or fname.find(".tgz") >= 0:
        with tarfile.open(fname) as tf:

            def is_within_directory(directory, target):

                abs_directory = os.path.abspath(directory)
                abs_target = os.path.abspath(target)

                prefix = os.path.commonprefix([abs_directory, abs_target])

                return prefix == abs_directory

            def safe_extract(tar, path=".", members=None, *, numeric_owner=False):

                for member in tar.getmembers():
                    member_path = os.path.join(path, member.name)
                    if not is_within_directory(path, member_path):
                        raise Exception("Attempted Path Traversal in Tar File")

                tar.extractall(path, members, numeric_owner=numeric_owner)

            safe_extract(tf, path=fpath_tmp)
    elif fname.find(".zip") >= 0:
        with zipfile.ZipFile(fname) as zf:
            zf.extractall(path=fpath_tmp)
    else:
        raise TypeError("Unsupport compress file type {}".format(fname))

    for f in os.listdir(fpath_tmp):
        src_dir = osp.join(fpath_tmp, f)
        dst_dir = osp.join(fpath, f)
        move_and_merge_tree(src_dir, dst_dir)

    shutil.rmtree(fpath_tmp)
    logging.debug("{} decompressed.".format(fname))
    return dst_dir


def url2dir(url, path, rename=None):
    full_name = download(url, path, rename, show_progress=True)
    print("File is downloaded, now extracting...")
    if url.count(".tgz") > 0 or url.count(".tar") > 0 or url.count("zip") > 0:
        return decompress(full_name)


def download_and_decompress(url, path=".", rename=None):
    fname = osp.split(url)[-1]
    fullname = osp.join(path, fname)
    # if url.endswith(('tgz', 'tar.gz', 'tar', 'zip')):
    #     fullname = osp.join(path, fname.split('.')[0])
    nranks = 0
    if nranks <= 1:
        dst_dir = url2dir(url, path, rename)
        if dst_dir is not None:
            fullname = dst_dir
    else:
        lock_path = fullname + ".lock"
        if not os.path.exists(fullname):
            with open(lock_path, "w"):
                os.utime(lock_path, None)
            if nranks == 0:
                dst_dir = url2dir(url, path, rename)
                if dst_dir is not None:
                    fullname = dst_dir
                os.remove(lock_path)
            else:
                while os.path.exists(lock_path):
                    time.sleep(1)
    return


def get_model_list(category: str = None):
    """
    Get all pre-trained models information supported by fd.download_model.
    Args:
        category(str): model category, if None, list all models in all categories.
    Returns:
        results(dict): a dictionary, key is category, value is a list which contains models information.
    """
    result = model_server.get_model_list()
    if result["status"] != 0:
        raise ValueError(
            "Failed to get pretrained models information from hub model server."
        )
    result = result["data"]
    if category is None:
        return result
    elif category in result:
        return {category: result[category]}
    else:
        raise ValueError(
            "No pretrained model in category {} can be downloaded now.".format(category)
        )


def download_model(
    name: str, path: str = None, format: str = None, version: str = None
):
    """
    Download pre-trained model for UltraInfer inference engine.
    Args:
        name: model name
        path(str): local path for saving model. If not set, default is hubenv.MODEL_HOME
        format(str): UltraInfer model format
        version(str) : UltraInfer model version
    """
    result = model_server.search_model(name, format, version)
    if path is None:
        path = hubenv.MODEL_HOME
    if result:
        url = result[0]["url"]
        format = result[0]["format"]
        version = result[0]["version"]
        fullpath = download(url, path, show_progress=True)
        model_server.stat_model(name, format, version)
        if format == "paddle":
            if url.count(".tgz") > 0 or url.count(".tar") > 0 or url.count("zip") > 0:
                archive_path = fullpath
                fullpath = decompress(fullpath)
                try:
                    os.rename(fullpath, os.path.join(os.path.dirname(fullpath), name))
                    fullpath = os.path.join(os.path.dirname(fullpath), name)
                    os.remove(archive_path)
                except FileExistsError:
                    pass
        print("Successfully download model at path: {}".format(fullpath))
        return fullpath
    else:
        print("ERROR: Could not find a model named {}".format(name))



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import platform
import sys


def process_paddle_inference(paddle_inference_so_file, paddle_inference_version):
    if platform.system().lower() != "linux":
        return
    rpaths = [
        "$ORIGIN",
        "$ORIGIN/../../third_party/install/mklml/lib/",
        "$ORIGIN/../../third_party/install/xpu/lib/",
        "$ORIGIN/../../third_party/install/fdmodel/lib/",
        "$ORIGIN/../../../tensorrt/lib/",
    ]
    version_major = int(paddle_inference_version.split(".")[0])
    if paddle_inference_version != "0.0.0" and version_major < 2:
        raise ValueError("Invalid Paddle Inference version")
    if version_major == 2:
        rpaths.append("$ORIGIN/../../third_party/install/mkldnn/lib/")
    else:
        rpaths.append("$ORIGIN/../../third_party/install/onednn/lib/")

    patchelf_exe = os.getenv("PATCHELF_EXE", "patchelf")
    command = "{} --force-rpath --set-rpath '{}' {}".format(
        patchelf_exe, ":".join(rpaths), paddle_inference_so_file
    )
    if platform.machine() != "sw_64" and platform.machine() != "mips64":
        assert os.system(command) == 0, "patchelf {} failed, the command: {}".format(
            paddle_inference_so_file, command
        )


if __name__ == "__main__":
    process_paddle_inference(sys.argv[1], sys.argv[2])



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import logging
from dataclasses import dataclass
from typing import List

import numpy as np

from .... import UltraInferModel, ModelFormat
from .... import c_lib_wrap as C
from ...common import ProcessorManager
from ....py_only import PyOnlyProcessorChain
from ....py_only.vision import PyOnlyVisionModel, processors as P
from ....utils.misc import load_config


class PaddleSegModel(UltraInferModel):
    def __init__(
        self,
        model_file,
        params_file,
        config_file,
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        """Load a image segmentation model exported by PaddleSeg.

        :param model_file: (str)Path of model file, e.g unet/model.pdmodel
        :param params_file: (str)Path of parameters file, e.g unet/model.pdiparams, if the model_fomat is ModelFormat.ONNX, this param will be ignored, can be set as empty string
        :param config_file: (str) Path of configuration file for deploy, e.g unet/deploy.yml
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model
        """
        super(PaddleSegModel, self).__init__(runtime_option)

        # assert model_format == ModelFormat.PADDLE, "PaddleSeg only support model format of ModelFormat.Paddle now."
        self._model = C.vision.segmentation.PaddleSegModel(
            model_file, params_file, config_file, self._runtime_option, model_format
        )
        assert self.initialized, "PaddleSeg model initialize failed."

    def predict(self, image):
        """Predict the segmentation result for an input image

        :param im: (numpy.ndarray)The input image data, 3-D array with layout HWC, BGR format
        :return: SegmentationResult
        """
        return self._model.predict(image)

    def batch_predict(self, image_list):
        """Predict the segmentation results for a batch of input images

        :param image_list: (list of numpy.ndarray) The input image list, each element is a 3-D array with layout HWC, BGR format
        :return: list of SegmentationResult
        """
        return self._model.batch_predict(image_list)

    def clone(self):
        """Clone PaddleSegModel object

        :return: a new PaddleSegModel object
        """

        class PaddleSegCloneModel(PaddleSegModel):
            def __init__(self, model):
                self._model = model

        clone_model = PaddleSegCloneModel(self._model.clone())
        return clone_model

    @property
    def preprocessor(self):
        """Get PaddleSegPreprocessor object of the loaded model

        :return: PaddleSegPreprocessor
        """
        return self._model.preprocessor

    @property
    def postprocessor(self):
        """Get PaddleSegPostprocessor object of the loaded model

        :return: PaddleSegPostprocessor
        """
        return self._model.postprocessor


class PaddleSegPreprocessor(ProcessorManager):
    def __init__(self, config_file):
        """Create a preprocessor for PaddleSegModel from configuration file

        :param config_file: (str)Path of configuration file, e.g ppliteseg/deploy.yaml
        """
        self._manager = C.vision.segmentation.PaddleSegPreprocessor(config_file)

    def disable_normalize(self):
        """
        This function will disable normalize in preprocessing step.
        """
        self._manager.disable_normalize()

    def disable_permute(self):
        """
        This function will disable hwc2chw in preprocessing step.
        """
        self._manager.disable_permute()

    @property
    def is_vertical_screen(self):
        """Attribute of PP-HumanSeg model. Stating Whether the input image is vertical image(height > width), default value is False

        :return: value of is_vertical_screen(bool)
        """
        return self._manager.is_vertical_screen

    @is_vertical_screen.setter
    def is_vertical_screen(self, value):
        """Set attribute is_vertical_screen of PP-HumanSeg model.

        :param value: (bool)The value to set is_vertical_screen
        """
        assert isinstance(
            value, bool
        ), "The value to set `is_vertical_screen` must be type of bool."
        self._manager.is_vertical_screen = value


class PaddleSegPostprocessor:
    def __init__(self, config_file):
        """Create a postprocessor for PaddleSegModel from configuration file

        :param config_file: (str)Path of configuration file, e.g ppliteseg/deploy.yaml
        """
        self._postprocessor = C.vision.segmentation.PaddleSegPostprocessor(config_file)

    def run(self, runtime_results, imgs_info):
        """Postprocess the runtime results for PaddleSegModel

        :param runtime_results: (list of FDTensor)The output FDTensor results from runtime
        :param imgs_info: The original input images shape info map, key is "shape_info", value is [[image_height, image_width]]
        :return: list of SegmentationResult(If the runtime_results is predict by batched samples, the length of this list equals to the batch size)
        """
        return self._postprocessor.run(runtime_results, imgs_info)

    @property
    def apply_softmax(self):
        """Attribute of PaddleSeg model. Stating Whether applying softmax operator in the postprocess, default value is False

        :return: value of apply_softmax(bool)
        """
        return self._postprocessor.apply_softmax

    @apply_softmax.setter
    def apply_softmax(self, value):
        """Set attribute apply_softmax of PaddleSeg model.

        :param value: (bool)The value to set apply_softmax
        """
        assert isinstance(
            value, bool
        ), "The value to set `apply_softmax` must be type of bool."
        self._postprocessor.apply_softmax = value

    @property
    def store_score_map(self):
        """Attribute of PaddleSeg model. Stating Whether storing score map in the SegmentationResult, default value is False

        :return: value of store_score_map(bool)
        """
        return self._postprocessor.store_score_map

    @store_score_map.setter
    def store_score_map(self, value):
        """Set attribute store_score_map of PaddleSeg model.

        :param value: (bool)The value to set store_score_map
        """
        assert isinstance(
            value, bool
        ), "The value to set `store_score_map` must be type of bool."
        self._postprocessor.store_score_map = value


class PyOnlyAnomalyDetectionModel(PyOnlyVisionModel):
    def __init__(
        self,
        model_file,
        params_file,
        config_file,
        runtime_option=None,
        model_format=ModelFormat.PADDLE,
    ):
        self._model_file = model_file
        self._params_file = params_file
        self._model_format = model_format
        super().__init__(runtime_option)
        self._config = load_config(config_file)
        self._preprocessor = _PyOnlyAnomalyDetectionPreprocessor(
            self._config["Deploy"]["transforms"]
        )
        self._postprocessor = _PyOnlyAnomalyDetectionPostprocessor()

    def model_name():
        return "PyOnlyImageAnomalyDetectionModel"

    def batch_predict(self, imgs):
        data_list = []
        for img in imgs:
            data = {"img": img}
            data = self._preprocessor.run(data)
            data_list.append(data)

        input_name = self._runtime.get_input_info(0).name
        imgs = np.stack([data["img"] for data in data_list], axis=0, dtype=np.float32)
        imgs = np.ascontiguousarray(imgs)
        output_arrs = self._runtime.infer({input_name: imgs})

        results = []
        for score_map in output_arrs[0]:
            data = {"score_map": score_map}
            result = self._postprocessor.run(data)
            results.append(result)
        return results

    def _update_option(self):
        self._option.set_model_path(
            self._model_file, self._params_file, self._model_format
        )


class _PyOnlyAnomalyDetectionPreprocessor(object):
    def __init__(self, config):
        super().__init__()
        processors = self._build_processors(config)
        processors.append(P.ToCHWImage())
        self._processor_chain = PyOnlyProcessorChain(processors)

    def run(self, data):
        return self._processor_chain(data)

    def _build_processors(self, config):
        processors = []
        for item in config:
            tf_type = item["type"]
            args = {k: v for k, v in item.items() if k != "type"}
            if tf_type == "Resize":
                if args.keys() - {
                    "target_size",
                    "keep_ratio",
                    "size_divisor",
                    "interp",
                }:
                    raise ValueError
                args.setdefault("keep_ratio", False)
                args.setdefault("size_divisor", None)
                args.setdefault("interp", "LINEAR")
                processor = P.Resize(
                    target_size=args["target_size"],
                    keep_ratio=args["keep_ratio"],
                    size_divisor=args["size_divisor"],
                    interp=args["interp"],
                )
            elif tf_type == "ResizeByLong":
                if args.keys() - {"long_size"}:
                    raise ValueError
                args.setdefault("size_divisor", None)
                args.setdefault("interp", "LINEAR")
                processor = P.ResizeByLong(target_long_edge=args["long_size"])
            elif tf_type == "ResizeByShort":
                if args.keys() - {"short_size"}:
                    raise ValueError
                processor = P.ResizeByShort(target_short_edge=args["short_size"])
            elif tf_type == "Normalize":
                if args.keys() - {"mean", "std"}:
                    raise ValueError
                args.setdefault("mean", 0.5)
                args.setdefault("std", 0.5)
                processor = P.Normalize(mean=args["mean"], std=args["std"])
            else:
                raise ValueError("Unknown transform type")
            processors.append(processor)
        return processors


class _PyOnlyAnomalyDetectionPostprocessor(object):
    def run(self, data):
        from skimage import morphology

        score_map = data["score_map"]

        thred = 0.01
        mask = score_map[0]
        mask[mask > thred] = 255
        mask[mask <= thred] = 0
        kernel = morphology.disk(4)
        mask = morphology.opening(mask, kernel)
        mask = mask.astype(np.uint8)

        result = _PyOnlyAnomalyDetectionResult(
            label_map=mask.reshape((-1)).tolist(), shape=list(mask.shape)
        )
        return result


@dataclass
class _PyOnlyAnomalyDetectionResult(object):
    label_map: List[int]
    shape: List[int]



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module is used to store environmental variables for ultra_infer model hub.

ULTRAINFER_HUB_HOME              -->  the root directory for storing ultra_infer model hub related data. Default to ~/.ultra_infer. Users can change the
├                          default value through the ULTRAINFER_HUB_HOME environment variable.
├── MODEL_HOME       -->   Store the downloaded ultra_infer models.
├── CONF_HOME        -->   Store the default configuration files.
"""

import os


def _get_user_home():
    return os.path.expanduser("~")


def _get_hub_home():
    if "ULTRAINFER_HUB_HOME" in os.environ:
        home_path = os.environ["ULTRAINFER_HUB_HOME"]
        if os.path.exists(home_path):
            if os.path.isdir(home_path):
                return home_path
            else:
                raise RuntimeError(
                    "The environment variable ULTRAINFER_HUB_HOME {} is not a directory.".format(
                        home_path
                    )
                )
        else:
            return home_path
    return os.path.join(_get_user_home(), ".ultra_infer")


def _get_sub_home(directory):
    home = os.path.join(_get_hub_home(), directory)
    os.makedirs(home, exist_ok=True)
    return home


USER_HOME = _get_user_home()
HUB_HOME = _get_hub_home()
MODEL_HOME = _get_sub_home("models")
CONF_HOME = _get_sub_home("conf")
RESOURCE_HOME = _get_sub_home("resources")



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ...utils.benchmark import benchmark


@benchmark.timeit
class NormalizeFeatures:
    """Normalize Features Transform"""

    def _normalize(self, preds):
        """normalize"""
        feas_norm = np.sqrt(np.sum(np.square(preds), axis=1, keepdims=True))
        features = np.divide(preds, feas_norm)
        return features

    def __call__(self, preds):
        return self._normalize(preds[0])



# Copyright (c) 2025 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Any, Dict, List, Optional, Union

import numpy as np

from ...models.text_to_pinyin.result import TextToPinyinResult
from ...models.text_to_speech_acoustic.result import Fastspeech2Result
from ...models.text_to_speech_vocoder.result import PwganResult
from ...utils.benchmark import benchmark
from ...utils.hpi import HPIConfig
from ...utils.pp_option import PaddlePredictorOption
from ..base import BasePipeline


@benchmark.time_methods
class TextToSpeechPipeline(BasePipeline):
    """Text to Speech Pipeline Pipeline"""

    entities = "text_to_speech"

    def __init__(
        self,
        config: Dict,
        device: str = None,
        pp_option: PaddlePredictorOption = None,
        use_hpip: bool = False,
        hpi_config: Optional[Union[Dict[str, Any], HPIConfig]] = None,
    ) -> None:
        """
        Initializes the class with given configurations and options.

        Args:
            config (Dict): Configuration dictionary containing model and other parameters.
            device (str): The device to run the prediction on. Default is None.
            pp_option (PaddlePredictorOption): Options for PaddlePaddle predictor. Default is None.
            use_hpip (bool, optional): Whether to use the high-performance
                inference plugin (HPIP) by default. Defaults to False.
            hpi_config (Optional[Union[Dict[str, Any], HPIConfig]], optional):
                The default high-performance inference configuration dictionary.
                Defaults to None.
        """
        super().__init__(
            device=device, pp_option=pp_option, use_hpip=use_hpip, hpi_config=hpi_config
        )

        text_to_pinyin_model_config = config["SubModules"]["TextToPinyin"]
        self.text_to_pinyin_model = self.create_model(text_to_pinyin_model_config)
        text_to_speech_acoustic_model_config = config["SubModules"][
            "TextToSpeechAcoustic"
        ]
        self.text_to_speech_acoustic_model = self.create_model(
            text_to_speech_acoustic_model_config
        )
        text_to_speech_vocoder_model_config = config["SubModules"][
            "TextToSpeechVocoder"
        ]
        self.text_to_speech_vocoder_model = self.create_model(
            text_to_speech_vocoder_model_config
        )

    def predict(
        self, input: Union[str, List[str], np.ndarray, List[np.ndarray]], **kwargs
    ) -> PwganResult:
        """Predicts speech recognition results for the given input.

        Args:
            input (Union[str, list[str], np.ndarray, list[np.ndarray]]): The input audio or path.
            **kwargs: Additional keyword arguments that can be passed to the function.

        Returns:
            PwganResult: The predicted pwgan results, support str and json output.
        """
        sentences = []
        if isinstance(input, str):
            if input.endswith(".txt"):
                if not os.path.exists(input):
                    raise FileNotFoundError(
                        f"The specified text file does not exist: {input}"
                    )
                try:
                    with open(input, "r", encoding="utf-8") as f:
                        sentences = [line.strip() for line in f.readlines()]
                except IOError as e:
                    raise IOError(
                        f"An error occurred while reading the file {input}: {e}"
                    )
            else:
                sentences = [input]
        elif isinstance(input, list):
            for item in input:
                if isinstance(item, str):
                    if item.endswith(".txt"):
                        if not os.path.exists(item):
                            raise FileNotFoundError(
                                f"The specified text file in the list does not exist: {item}"
                            )
                        try:
                            with open(item, "r", encoding="utf-8") as f:
                                sentences.extend(
                                    [line.strip() for line in f.readlines()]
                                )
                        except IOError as e:
                            raise IOError(
                                f"An error occurred while reading the file {item}: {e}"
                            )
                    else:
                        sentences.append(item)
        else:
            raise TypeError(
                f"Unsupported input type: {type(input)}. Expected str, list, or np.ndarray."
            )
        if not sentences:
            raise ValueError(
                "The input resulted in an empty list of sentences to process."
            )

        for sentence in sentences:
            text_to_pinyin_res = [
                self.get_text_to_pinyin_result(sentence)["result"]["phone_ids"]
            ]
            text_to_speech_acoustic_res = [
                self.get_text_to_speech_acoustic_result(text_to_pinyin_res)["result"]
            ]
            yield from self.text_to_speech_vocoder_model(text_to_speech_acoustic_res)

    def get_text_to_pinyin_result(
        self, input: Union[str, List[str]]
    ) -> TextToPinyinResult:
        """Get the result of text to pinyin conversion.

        Args:
            input (Union[str, list[str]]): The input text or list of texts.

        Returns:
            TextToPinyinResult: The result of text to pinyin conversion.
        """
        return next(self.text_to_pinyin_model(input))

    def get_text_to_speech_acoustic_result(
        self, input: Union[str, List[str]]
    ) -> Fastspeech2Result:
        """Get the result of text to speech acoustic conversion.

        Args:
            input (Union[str, list[str]]): The input text or list of texts.

        Returns:
            Fastspeech2Result: The result of text to speech acoustic conversion.
        """
        return next(self.text_to_speech_acoustic_model(input))



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import

from . import (
    classification,
    detection,
    evaluation,
    facealign,
    facedet,
    faceid,
    generation,
    headpose,
    keypointdetection,
    matting,
    ocr,
    perception,
    segmentation,
    sr,
    tracking,
)
from .utils import fd_result_to_json
from .visualize import *
from .. import C


def enable_flycv():
    return C.vision.enable_flycv()


def disable_flycv():
    return C.vision.disable_flycv()



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from .... import UltraInferModel, ModelFormat
from .... import c_lib_wrap as C


class Yolov7FacePreprocessor:
    def __init__(self):
        """Create a preprocessor for Yolov7Face"""
        self._preprocessor = C.vision.facedet.Yolov7Preprocessor()

    def run(self, input_ims):
        """Preprocess input images for Yolov7Face

        :param: input_ims: (list of numpy.ndarray)The input image
        :return: list of FDTensor
        """
        return self._preprocessor.run(input_ims)

    @property
    def size(self):
        """
        Argument for image preprocessing step, the preprocess image size, tuple of (width, height), default size = [640, 640]
        """
        return self._preprocessor.size

    @property
    def padding_color_value(self):
        """
        padding value for preprocessing, default [114.0, 114.0, 114.0]
        """
        #  padding value, size should be the same as channels
        return self._preprocessor.padding_color_value

    @property
    def is_scale_up(self):
        """
        is_scale_up for preprocessing, the input image only can be zoom out, the maximum resize scale cannot exceed 1.0, default true
        """
        return self._preprocessor.is_scale_up

    @size.setter
    def size(self, wh):
        assert isinstance(
            wh, (list, tuple)
        ), "The value to set `size` must be type of tuple or list."
        assert (
            len(wh) == 2
        ), "The value to set `size` must contains 2 elements means [width, height], but now it contains {} elements.".format(
            len(wh)
        )
        self._preprocessor.size = wh

    @padding_color_value.setter
    def padding_color_value(self, value):
        assert isinstance(
            value, list
        ), "The value to set `padding_color_value` must be type of list."
        self._preprocessor.padding_color_value = value

    @is_scale_up.setter
    def is_scale_up(self, value):
        assert isinstance(
            value, bool
        ), "The value to set `is_scale_up` must be type of bool."
        self._preprocessor.is_scale_up = value


class Yolov7FacePostprocessor:
    def __init__(self):
        """Create a postprocessor for Yolov7Face"""
        self._postprocessor = C.vision.facedet.Yolov7FacePostprocessor()

    def run(self, runtime_results, ims_info):
        """Postprocess the runtime results for Yolov7Face

        :param: runtime_results: (list of FDTensor)The output FDTensor results from runtime
        :param: ims_info: (list of dict)Record input_shape and output_shape
        :return: list of DetectionResult(If the runtime_results is predict by batched samples, the length of this list equals to the batch size)
        """
        return self._postprocessor.run(runtime_results, ims_info)

    @property
    def conf_threshold(self):
        """
        confidence threshold for postprocessing, default is 0.5
        """
        return self._postprocessor.conf_threshold

    @property
    def nms_threshold(self):
        """
        nms threshold for postprocessing, default is 0.45
        """
        return self._postprocessor.nms_threshold

    @property
    def landmarks_per_face(self):
        """
        landmarks per face for postprocessing, default is 5
        """
        return self._postprocessor.landmarks_per_face

    @conf_threshold.setter
    def conf_threshold(self, conf_threshold):
        assert isinstance(
            conf_threshold, float
        ), "The value to set `conf_threshold` must be type of float."
        self._postprocessor.conf_threshold = conf_threshold

    @nms_threshold.setter
    def nms_threshold(self, nms_threshold):
        assert isinstance(
            nms_threshold, float
        ), "The value to set `nms_threshold` must be type of float."
        self._postprocessor.nms_threshold = nms_threshold

    @landmarks_per_face.setter
    def landmarks_per_face(self, landmarks_per_face):
        assert isinstance(
            landmarks_per_face, int
        ), "The value to set `landmarks_per_face` must be type of int."
        self._postprocessor.landmarks_per_face = landmarks_per_face


class YOLOv7Face(UltraInferModel):
    def __init__(
        self,
        model_file,
        params_file="",
        runtime_option=None,
        model_format=ModelFormat.ONNX,
    ):
        """Load a YOLOv7Face model exported by YOLOv7Face.

        :param model_file: (str)Path of model file, e.g ./yolov7face.onnx
        :param params_file: (str)Path of parameters file, e.g yolox/model.pdiparams, if the model_fomat is ModelFormat.ONNX, this param will be ignored, can be set as empty string
        :param runtime_option: (ultra_infer.RuntimeOption)RuntimeOption for inference this model, if it's None, will use the default backend on CPU
        :param model_format: (ultra_infer.ModelForamt)Model format of the loaded model
        """
        super(YOLOv7Face, self).__init__(runtime_option)

        self._model = C.vision.facedet.YOLOv7Face(
            model_file, params_file, self._runtime_option, model_format
        )

        assert self.initialized, "YOLOv7Face initialize failed."

    def predict(self, input_image):
        """Detect the location and key points of human faces from an input image
        :param input_image: (numpy.ndarray)The input image data, 3-D array with layout HWC, BGR format
        :return: FaceDetectionResult
        """
        return self._model.predict(input_image)

    def batch_predict(self, images):
        """Classify a batch of input image

        :param im: (list of numpy.ndarray) The input image list, each element is a 3-D array with layout HWC, BGR format
        :return list of DetectionResult
        """

        return self._model.batch_predict(images)

    @property
    def preprocessor(self):
        """Get YOLOv7Preprocessor object of the loaded model

        :return YOLOv7Preprocessor
        """
        return self._model.preprocessor

    @property
    def postprocessor(self):
        """Get YOLOv7Postprocessor object of the loaded model

        :return YOLOv7Postprocessor
        """
        return self._model.postprocessor
